# coding: utf-8

from __future__ import unicode_literals

from threading import Lock
import random
import string  # pylint:disable=deprecated-module

from six.moves.urllib.parse import urlencode, urlunsplit  # pylint:disable=import-error,no-name-in-module

from boxsdk.network.default_network import DefaultNetwork
from boxsdk.config import API
from boxsdk.exception import BoxOAuthException


class OAuth2(object):
    """
    Responsible for handling OAuth2 for the Box API. Can authenticate and refresh tokens.
    """

    def __init__(
            self,
            client_id,
            client_secret,
            store_tokens=None,
            box_device_id='0',
            box_device_name='',
            access_token=None,
            refresh_token=None,
            network_layer=None,
    ):
        """
        :param client_id:
            Box API key used for identifying the application the user is authenticating with.
        :type client_id:
            `unicode`
        :param client_secret:
            Box API secret used for making OAuth2 requests.
        :type client_secret:
            `unicode`
        :param store_tokens:
            Optional callback for getting access to tokens for storing them.
        :type store_tokens:
            `callable`
        :param box_device_id:
            Optional unique ID of this device. Used for applications that want to support device-pinning.
        :type box_device_id:
            `unicode`
        :param box_device_name:
            Optional human readable name for this device.
        :type box_device_name:
            `unicode`
        :param access_token:
            Access token to use for auth until it expires.
        :type access_token:
            `unicode`
        :param refresh_token:
            Refresh token to use for auth until it expires or is used.
        :type refresh_token:
            `unicode`
        :param network_layer:
            If specified, use it to make network requests. If not, the default network implementation will be used.
        :type network_layer:
            :class:`Network`
        """
        self._client_id = client_id
        self._client_secret = client_secret
        self._store_tokens = store_tokens
        self._access_token = access_token
        self._refresh_token = refresh_token
        self._network_layer = network_layer if network_layer else DefaultNetwork()
        self._refresh_lock = Lock()
        self._box_device_id = box_device_id
        self._box_device_name = box_device_name

    @property
    def access_token(self):
        """
        Get the current access token.

        :return:
            current access token
        :rtype:
            `unicode`
        """
        return self._access_token

    def get_authorization_url(self, redirect_url):
        """
        Get the authorization url based on the client id and the redirect url, passed in

        :param redirect_url:
            An HTTPS URI or custom URL scheme where the response will be redirected. Optional if the redirect URI is
            registered with Box already.
        :type redirect_url:
            `unicode` or None
        :return:
            A tuple of the URL of Box’s authorization page and the CSRF token.
            This is the URL that your application should forward the user to in first leg of OAuth 2.
        :rtype:
            (`unicode`, `unicode`)
        """
        csrf_token = self._get_state_csrf_token()
        # For the query string parameters, use a sequence of two-element
        # tuples, rather than a dictionary, in order to get a consistent and
        # predictable order of parameters in the output of `urlencode()`.
        params = [
            ('state', csrf_token),
            ('response_type', 'code'),
            ('client_id', self._client_id),
        ]
        if redirect_url:
            params.append(('redirect_uri', redirect_url))
        # `urlencode()` doesn't work with non-ASCII unicode characters, so
        # encode the parameters as ASCII bytes.
        params = [(key.encode('utf-8'), value.encode('utf-8')) for (key, value) in params]
        query_string = urlencode(params)
        return urlunsplit(('', '', API.OAUTH2_AUTHORIZE_URL, query_string, '')), csrf_token

    def authenticate(self, auth_code):
        """
        Send token request and return the access_token, refresh_token tuple. The access token and refresh token will be
        stored by calling the `store_tokens` callback if provided in __init__.

        :param auth_code:
            An authorization code you retrieved in the first leg of OAuth 2.
        :type auth_code:
            `unicode` or None

        :return:
            (access_token, refresh_token)
        :rtype:
            (`unicode`, `unicode`)
        """
        data = {
            'grant_type': 'authorization_code',
            'code': auth_code,
            'client_id': self._client_id,
            'client_secret': self._client_secret,
        }
        if self._box_device_id:
            data['box_device_id'] = self._box_device_id
        if self._box_device_name:
            data['box_device_name'] = self._box_device_name
        return self.send_token_request(data, access_token=None)

    def _refresh(self, access_token):
        data = {
            'grant_type': 'refresh_token',
            'refresh_token': self._refresh_token,
            'client_id': self._client_id,
            'client_secret': self._client_secret,
        }
        if self._box_device_id:
            data['box_device_id'] = self._box_device_id
        if self._box_device_name:
            data['box_device_name'] = self._box_device_name

        return self.send_token_request(data, access_token)

    def refresh(self, access_token_to_refresh):
        """
        Refresh the access token and the refresh token and return the access_token, refresh_token tuple. The access
        token and refresh token will be stored by calling the `store_tokens` callback if provided in __init__.

        :param access_token_to_refresh:
            The expired access token, which needs to be refreshed.
        :type access_token_to_refresh:
            `unicode`
        """
        with self._refresh_lock:
            # The lock here is for handling that case that multiple requests fail, due to access token expired, at the
            # same time to avoid multiple session renewals.
            if access_token_to_refresh == self._access_token:
                # If the active access token is the same as the token needs to be refreshed, we make the request to
                # refresh the token.
                return self._refresh(access_token_to_refresh)
            else:
                # If the active access token (self._access_token) is not the same as the token needs to be refreshed,
                # it means the expired token has already been refreshed. Simply return the current active tokens.
                return self._access_token, self._refresh_token

    @staticmethod
    def _get_state_csrf_token():
        """ Generate a random state CSRF token to be used in the authorization url.
        Example: box_csrf_token_Iijw9aU31sNdgiQu

        :return:
            The security token
        :rtype:
            `unicode`
        """
        system_random = random.SystemRandom()
        ascii_alphabet = string.ascii_letters + string.digits
        ascii_len = len(ascii_alphabet)
        return 'box_csrf_token_' + ''.join(ascii_alphabet[int(system_random.random() * ascii_len)] for _ in range(16))

    def send_token_request(self, data, access_token):
        """
        Send the request to acquire or refresh an access token.

        :param data:
            Dictionary containing the request parameters as specified by the Box API.
        :type data:
            `dict`
        :param access_token:
            The current access token.
        :type access_token:
            `unicode` or None
        :return:
            The access token and refresh token.
        :rtype:
            (`unicode`, `unicode`)
        """
        url = '{base_auth_url}/token'.format(base_auth_url=API.OAUTH2_API_URL)
        headers = {'content-type': 'application/x-www-form-urlencoded'}
        network_response = self._network_layer.request(
            'POST',
            url,
            data=data,
            headers=headers,
            access_token=access_token
        )
        if not network_response.ok:
            raise BoxOAuthException(network_response.status_code, network_response.content, url, 'POST')
        try:
            response = network_response.json()
            self._access_token = response['access_token']
            self._refresh_token = response['refresh_token']
        except (ValueError, KeyError):
            raise BoxOAuthException(network_response.status_code, network_response.content, url, 'POST')
        if self._store_tokens:
            self._store_tokens(self._access_token, self._refresh_token)
        return self._access_token, self._refresh_token



# coding: utf-8

from __future__ import unicode_literals
from bottle import request, response
from sqlalchemy.orm import make_transient
from sqlalchemy.orm.exc import NoResultFound
from boxsdk.object.folder import FolderSyncState
from test.functional.mock_box.behavior.item_behavior import ItemBehavior
from test.functional.mock_box.db_model.event_model import EventModel
from test.functional.mock_box.db_model.file_model import FileModel
from test.functional.mock_box.db_model.folder_model import FolderModel
from test.functional.mock_box.util.db_utils import get_folder_by_id, get_user_from_header
from test.functional.mock_box.util.http_utils import abort
from test.functional.mock_box.util import json_utils as json


class FolderBehavior(ItemBehavior):
    def get_folder_info(self, folder_id):
        """
        https://developers.box.com/docs/#folders-get-information-about-a-folder
        """
        folder = get_folder_by_id(self._db_session, folder_id)
        return json.dumps(folder)

    def update_folder_info(self, folder_id):
        folder = get_folder_by_id(self._db_session, folder_id)
        params = json.load(request.body)
        for key, value in params.items():
            if not hasattr(FolderModel, key):
                abort(400, 'Folder has no attribute {0}.'.format(key))
            if key == 'parent':
                # Move
                parent_id = value['id']
                parent_folder = get_folder_by_id(self._db_session, parent_id)
                folder.parent_id = parent_folder.id
                self._db_session.add(
                    EventModel(event_type='ITEM_MOVE', source_id=folder.folder_id, source_type='folder'),
                )
            else:
                setattr(folder, key, value)
                if key == 'name':
                    self._db_session.add(
                        EventModel(event_type='ITEM_RENAME', source_id=folder.folder_id, source_type='folder'),
                    )
                elif key == 'sync_state':
                    event_type = 'ITEM_SYNC' if value == FolderSyncState.IS_SYNCED else 'ITEM_UNSYNC'
                    self._db_session.add(
                        EventModel(event_type=event_type, source_id=folder.folder_id, source_type='folder'),
                    )
        self._db_session.commit()
        return json.dumps(folder)

    def copy_folder(self, folder_id):
        folder = get_folder_by_id(self._db_session, folder_id)
        parent_folder = self._get_parent()
        self._db_session.expunge(folder)
        make_transient(folder)
        folder.id = None
        folder.folder_id = None
        folder.parent_id = parent_folder.id
        self._db_session.add(folder)
        self._db_session.commit()
        self._db_session.add(EventModel(event_type='ITEM_COPY', source_id=folder.folder_id, source_type='folder'))
        self._db_session.commit()
        return json.dumps(folder)

    def delete_folder(self, folder_id):
        folder = get_folder_by_id(self._db_session, folder_id)
        self._db_session.delete(folder)
        self._db_session.commit()
        self._db_session.add(EventModel(event_type='ITEM_TRASH', source_id=folder.folder_id, source_type='folder'))
        response.status = 204

    def _gat(self, folder_id):
        # pylint:disable=unused-argument
        files = self._db_session.query(FileModel).all()
        folders = self._db_session.query(FolderModel).filter(FolderModel.folder_id != '0').all()
        return json.dumps({'items': files + folders})

    def get_folder_items(self, folder_id):
        view = request.params.get('view', None)
        if view is not None and view == 'subfolder_sync_forked_tree':
            return self._gat(folder_id)
        limit = int(request.params.get('limit', 100))
        offset = int(request.params.get('offset', 0))
        # fields = request.params.get('fields')
        # fields = set(fields.split(',')) if fields else set() | set(['name'])
        folder = get_folder_by_id(self._db_session, folder_id)
        folder_count = self._db_session.query(FolderModel).filter_by(parent_id=folder.id).count()
        folders = []
        if folder_count > offset:
            folders = self._db_session.query(
                FolderModel,
            ).filter_by(parent_id=folder.id).limit(limit).offset(offset).all()
        files = []
        if len(folders) < limit:
            limit -= len(folders)
            offset -= folder_count
            files = self._db_session.query(FileModel).filter_by(parent_id=folder.id).limit(limit).offset(offset).all()
        items = folders + files
        return json.dumps({'total_count': len(items), 'entries': items})

    def create_folder(self):
        """
        https://developers.box.com/docs/#folders-create-a-new-folder
        """
        params = json.load(request.body)
        name = params.get('name')
        if name is None:
            abort(400, 'Missing parameter: name')
        parent = params.get('parent')
        if parent is None or 'id' not in parent:
            abort(400, 'Missing parameter: parent(id)')
        parent_id = parent['id']
        try:
            parent_folder = self._db_session.query(FolderModel).filter_by(folder_id=parent_id).one()
        except NoResultFound:
            abort(404)
        owner = get_user_from_header(self._db_session)
        name_in_use = self._db_session.query(FolderModel).filter_by(name=name).count()
        if name_in_use:
            abort(409, 'An item with that name already exists.')
        folder = FolderModel(name=name, parent_id=parent_folder.id, owned_by=owner, created_by=owner)
        self._db_session.add(folder)
        self._db_session.commit()
        self._db_session.add(EventModel(event_type='ITEM_CREATE', source_id=folder.folder_id, source_type='folder'))
        self._db_session.commit()
        return json.dumps(folder)



# coding: utf-8

from __future__ import unicode_literals
from boxsdk.config import API


class BaseEndpoint(object):
    """A Box API endpoint."""

    def __init__(self, session):
        """

        :param session:
            The Box session used to make requests.
        :type session:
            :class:`BoxSession`
        """
        self._session = session

    def get_url(self, endpoint, *args):
        """
        Return the URL used to access the endpoint.

        :param endpoint:
            The name of the endpoint.
        :type endpoint:
            `url`
        :param args:
            Additional parts of the endpoint URL.
        :type args:
            `Iterable`
        :rtype:
            `unicode`
        """
        # pylint:disable=no-self-use
        url = ['{0}/{1}'.format(API.BASE_API_URL, endpoint)]
        url.extend(['/{0}'.format(x) for x in args])
        return ''.join(url)



# coding: utf-8

from __future__ import unicode_literals
from mock import DEFAULT, Mock, patch
import pytest
from requests import Response, Session
from boxsdk.network.default_network import DefaultNetworkResponse, DefaultNetwork


@pytest.fixture()
def mock_request(monkeypatch):
    mock_session_factory = Mock()
    mock_session_factory.return_value = session = Mock(Session, request=Mock())
    monkeypatch.setattr('requests.Session', mock_session_factory)
    return session.request


@pytest.fixture(params=('GET', 'POST', 'PUT', 'DELETE', 'OPTIONS'))
def http_verb(request):
    return request.param


def test_default_network_response_properties_pass_through_to_session_response_properties(access_token):
    mock_session_response = Mock(Response)
    mock_session_response.status_code = 200
    mock_session_response.headers = {}
    mock_session_response.raw = Mock()
    network_reponse = DefaultNetworkResponse(mock_session_response, access_token)
    assert network_reponse.json() == mock_session_response.json()
    assert network_reponse.content == mock_session_response.content
    assert network_reponse.ok == mock_session_response.ok
    assert network_reponse.status_code == mock_session_response.status_code
    assert network_reponse.headers == mock_session_response.headers
    assert network_reponse.response_as_stream == mock_session_response.raw
    assert network_reponse.access_token_used == access_token


def test_default_network_request(mock_request, http_verb, test_url, access_token, generic_successful_response):
    # pylint:disable=redefined-outer-name
    default_network = DefaultNetwork()
    mock_request.return_value = generic_successful_response
    response = default_network.request(http_verb, test_url, access_token, custom_kwarg='test')
    mock_request.assert_called_once_with(http_verb, test_url, custom_kwarg='test')
    assert response.response_as_stream == generic_successful_response.raw
    assert response.access_token_used == access_token


@pytest.mark.parametrize('delay', (0, 1))
def test_default_network_retry_after_sleeps(delay):
    default_network = DefaultNetwork()
    retry_call = Mock()
    mock_sleep = Mock()
    with patch('boxsdk.network.default_network.time.sleep', mock_sleep):
        default_network.retry_after(delay, retry_call, DEFAULT, kwarg=DEFAULT)
    mock_sleep.assert_called_once_with(delay)
    retry_call.assert_called_once_with(DEFAULT, kwarg=DEFAULT)



# coding: utf-8

from __future__ import unicode_literals
from boxsdk.util.ordered_dict import OrderedDict


class LRUCache(object):
    def __init__(self, capacity=512):
        """
        :param capacity:
            The Maximum number of key-value pairs can be cached.
        :type capacity:
            `int`
        """
        super(LRUCache, self).__init__()
        self.capacity = capacity
        self.cache = OrderedDict()

    def get(self, key):
        """
        Look up the value in cache using the associated key. Returns the value if found.
        Raises :class:`KeyError` otherwise.

        :param key:
            The key used to look up the cache.
        :type key:
            `unicode`
        :return:
            The value associated with the key if exists.
        :raises:
            :class:`KeyError` if the key doesn't exist.
        """
        value = self.cache.pop(key)
        self.cache[key] = value
        return value

    def set(self, key, value=None):
        """
        Store the key-value pair to cache.

        :param key:
            The key associated with the value to be stored. It's used to look up the cache.
        :type key:
            `unicode`
        :param value:
            The value to be stored.
        :type value:
            varies
        """
        try:
            self.cache.pop(key)
        except KeyError:
            if len(self.cache) >= self.capacity:
                self.cache.popitem(last=False)
        self.cache[key] = value



# coding: utf-8

from __future__ import unicode_literals
from bottle import request, redirect
from datetime import datetime, timedelta
import json
from sqlalchemy.orm.exc import NoResultFound
from uuid import uuid4
from test.functional.mock_box.db_model.application_model import ApplicationModel
from test.functional.mock_box.db_model.token_model import TokenModel
from test.functional.mock_box.db_model.user_model import UserModel
from test.functional.mock_box.util.db_utils import get_token_record_by_token
from test.functional.mock_box.util.http_utils import abort


class OAuth2Behavior(object):
    ACCESS_TOKEN_DURATION_SECONDS = 3600
    REFRESH_TOKEN_DURATION_DAYS = 60
    AUTH_CODE_DURATION_SECONDS = 30

    def __init__(self, db_session):
        self._db_session = db_session
        self._auth_request = {}

    def add_application(self, client_id, client_secret, user_ids):
        users = self._db_session.query(UserModel).filter(UserModel.user_id.in_(user_ids)).all()
        app = ApplicationModel(client_id=client_id, client_secret=client_secret, users=users)
        self._db_session.add(app)
        self._db_session.commit()

    def _get_application_by_id(self, client_id):
        try:
            return self._db_session.query(ApplicationModel).filter_by(client_id=client_id).one()
        except NoResultFound:
            abort(400, 'Invalid client id: {0}'.format(client_id))

    def _get_user_by_login(self, user_login):
        try:
            return self._db_session.query(UserModel).filter_by(login=user_login).one()
        except NoResultFound:
            abort(401)

    def check_authorization_header(self):
        """
        Check that the request has an auth header and that its token matches the currently valid token.
        Further check that the token isn't expired.

        Called by methods decorated with the authorize decorator.
        """
        authorization_header = request.headers.get('Authorization')
        if not authorization_header or not authorization_header.startswith('Bearer '):
            abort(401)
        token = authorization_header[7:]
        token_record = get_token_record_by_token(self._db_session, token)
        if datetime.utcnow() > token_record.expires_at:
            abort(401)

    def _create_tokens(self, client_id, user_login=None, owned_by_id=None):
        if owned_by_id is None:
            user = self._get_user_by_login(user_login)
            owned_by_id = user.id
        app = self._get_application_by_id(client_id)
        access_token, refresh_token = uuid4().hex, uuid4().hex
        access_token_valid_until = datetime.utcnow() + timedelta(seconds=self.ACCESS_TOKEN_DURATION_SECONDS)
        refresh_token_valid_until = datetime.utcnow() + timedelta(days=self.REFRESH_TOKEN_DURATION_DAYS)
        self._db_session.add(TokenModel(
            token=access_token,
            expires_at=access_token_valid_until,
            authorized_application_id=app.id,
            owned_by_id=owned_by_id,
            token_type='access'
        ))
        self._db_session.add(TokenModel(
            token=refresh_token,
            expires_at=refresh_token_valid_until,
            authorized_application_id=app.id,
            owned_by_id=owned_by_id,
            token_type='refresh'
        ))
        return access_token, access_token_valid_until, refresh_token, refresh_token_valid_until

    def oauth2_authorize(self):
        """
        Shortcut OAuth2 authorize method. Instead of presenting a webview for a user to login, provides the code
        directly. Saves the tokens so they can be issued by calls to token.
        """
        code = uuid4().hex
        state = request.params.state
        client_id = request.params.client_id
        redirect_uri = request.params.redirect_uri
        user_login = request.params.get('box_login', '')
        self._auth_request = {
            'code': code,
            'client_id': client_id,
            'redirect_uri': redirect_uri,
            'state': state,
        }
        return {'user_login': user_login, 'action': '{0}://{1}{2}'.format(*request.urlparts[:3])}

    def oauth2_finish_loop(self):
        user_login = request.forms.get('login')
        client_id = self._auth_request['client_id']
        access_token, _, refresh_token, _ = self._create_tokens(client_id, user_login)
        self._auth_request['access_token'] = access_token
        self._auth_request['refresh_token'] = refresh_token
        redirect_uri = self._auth_request['redirect_uri']
        code = self._auth_request['code']
        state = self._auth_request['state']
        self._db_session.commit()
        redirect('{0}?{1}'.format(redirect_uri, '&'.join(('code={0}'.format(code), 'state={0}'.format(state)))))

    def oauth2_token(self):
        """
        OAuth2 /token method.
        Either exchanges an auth code for an access/refresh token pair, or refreshes a token.
        """
        grant_type = request.forms.get('grant_type')
        client_id, client_secret = request.forms.get('client_id'), request.forms.get('client_secret')
        app = self._get_application_by_id(client_id)
        if client_secret != app.client_secret:
            abort(400, 'Invalid client secret: {0}'.format(client_secret))

        if grant_type == 'authorization_code':
            code = request.forms.get('code')
            if self._auth_request is None:
                abort(400, 'Invalid code: {0}'.format(code))
            access_token, refresh_token = self._auth_request['access_token'], self._auth_request['refresh_token']
        elif grant_type == 'refresh_token':
            refresh_token = request.forms.get('refresh_token')
            refresh_token_record = get_token_record_by_token(self._db_session, refresh_token)
            if refresh_token_record.token_type == 'refresh':
                if datetime.utcnow() > refresh_token_record.expires_at:
                    abort(400, 'Token expired: {0}'.format(refresh_token))
                access_token, _, refresh_token, _ = self._create_tokens(
                    client_id,
                    owned_by_id=refresh_token_record.owned_by_id,
                )
            else:
                abort(400, 'Invalid token: {0}'.format(refresh_token))
        else:
            abort(400, 'Invalid grant type: {0}'.format(grant_type))

        return json.dumps({
            'access_token': access_token,
            'refresh_token': refresh_token,
            'expires_in': self.ACCESS_TOKEN_DURATION_SECONDS,
        })

    def expire_token(self, token):
        token_record = get_token_record_by_token(self._db_session, token)
        token_record.expires_at = datetime.utcnow()
        self._db_session.commit()



# coding: utf-8

from __future__ import unicode_literals

from boxsdk.object.base_object import BaseObject
from boxsdk.util.text_enum import TextEnum


class CollaborationRole(TextEnum):
    """An enum of possible collaboration roles"""
    EDITOR = 'editor'
    VIEWER = 'viewer'

    # Available to enterprise accounts:
    PREVIEWER = 'previewer'
    UPLOADER = 'uploader'
    PREVIEWER_UPLOADER = 'previewer uploader'
    VIEWER_UPLOADER = 'viewer uploader'
    CO_OWNER = 'co-owner'
    OWNER = 'owner'


class CollaborationStatus(TextEnum):
    """An enum of possible statuses of a collaboration"""
    PENDING = 'pending'
    ACCEPTED = 'accepted'
    REJECTED = 'rejected'


class Collaboration(BaseObject):
    """An object that represents a collaboration between a folder and an individual or group"""
    _item_type = 'collaboration'

    # pylint:disable=arguments-differ
    def update_info(self, role=None, status=None):
        """Edit an existing collaboration on Box

        :param role:
            The new role for this collaboration or None to leave unchanged
        :type role:
            :class:`CollaborationRole`
        :param status:
            The new status for this collaboration or None to leave unchanged. A pending collaboration can be set to
            accepted or rejected if permissions allow it.
        :type status:
            :class:`CollaborationStatus`
        :returns:
            Whether or not the edit was successful.
        :rtype:
            `bool`
        :raises:
            :class:`BoxAPIException` if current user doesn't have permissions to edit the collaboration.
        """
        data = {}
        if role:
            data['role'] = role
        if status:
            data['status'] = status
        return super(Collaboration, self).update_info(data=data)



# coding: utf-8

from __future__ import unicode_literals
import json

from mock import Mock
import pytest
from six import text_type

# pylint:disable=redefined-builtin
# pylint:disable=import-error
from six.moves import zip
# pylint:enable=redefined-builtin
# pylint:enable=import-error

from boxsdk.auth.oauth2 import OAuth2
from boxsdk.client import Client
from boxsdk.config import API
from boxsdk.network.default_network import DefaultNetworkResponse
from boxsdk.object.events import Events
from boxsdk.object.folder import Folder
from boxsdk.object.file import File
from boxsdk.object.group import Group
from boxsdk.object.user import User
from boxsdk.object.group_membership import GroupMembership


@pytest.fixture()
def mock_client(mock_box_session):
    mock_oauth = Mock(OAuth2)
    client = Client(mock_oauth)
    # pylint:disable=protected-access
    client._session = mock_box_session
    return client


@pytest.fixture(scope='module')
def user_id_1():
    return 1


@pytest.fixture(scope='module')
def user_id_2():
    return 1023


@pytest.fixture(scope='module')
def file_id():
    return 100


@pytest.fixture(scope='module')
def folder_id():
    return '1022'


@pytest.fixture(scope='module')
def users_response(user_id_1, user_id_2):
    # pylint:disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.json.return_value = {'entries': [
        {'type': 'user', 'id': user_id_1}, {'type': 'user', 'id': user_id_2}
    ]}
    return mock_network_response


@pytest.fixture(scope='module')
def user_response(user_id_1):
    # pylint:disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.json.return_value = {'type': 'user', 'id': user_id_1}
    return mock_network_response


@pytest.fixture(scope='module')
def group_id_1():
    return 101


@pytest.fixture(scope='module')
def group_id_2():
    return 202


@pytest.fixture(scope='module')
def groups_response(group_id_1, group_id_2):
    # pylint:disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.json.return_value = {'entries': [
        {'type': 'group', 'id': group_id_1, 'name': text_type(group_id_1)},
        {'type': 'group', 'id': group_id_2, 'name': text_type(group_id_2)},
    ]}
    return mock_network_response


@pytest.fixture(scope='module')
def create_group_response():
    # pylint:disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.json.return_value = {
        'type': 'group',
        'id': 1234,
        'name': 'test_group_name',
    }
    return mock_network_response


@pytest.fixture(scope='module', params=('file', 'folder'))
def shared_item_response(request):
    # pylint:disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.json.return_value = {
        'type': request.param,
        'id': 1234,
        'name': 'shared_item',
    }
    return mock_network_response


@pytest.fixture(scope='module')
def search_response(file_id, folder_id):
    # pylint:disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.json.return_value = {'entries': [
        {'type': 'file', 'id': file_id}, {'type': 'folder', 'id': folder_id}
    ]}
    return mock_network_response


@pytest.mark.parametrize('test_class, factory_method_name', [
    (Folder, 'folder'),
    (File, 'file'),
    (User, 'user'),
    (Group, 'group'),
    (GroupMembership, 'group_membership'),
])
def test_factory_returns_the_correct_object(mock_client, test_class, factory_method_name):
    """ Tests the various id-only factory methods in the Client class """
    # pylint:disable=redefined-outer-name
    fake_id = 'fake_id'

    factory_method = getattr(mock_client, factory_method_name)

    obj = factory_method(fake_id)

    assert isinstance(obj, test_class)
    assert obj.object_id == fake_id


def test_users_return_the_correct_user_objects(mock_client, mock_box_session, users_response, user_id_1, user_id_2):
    # pylint:disable=redefined-outer-name
    mock_box_session.get.return_value = users_response
    users = mock_client.users()
    assert users[0].object_id == user_id_1
    assert users[1].object_id == user_id_2


def test_search_instantiates_search_and_calls_search(
        mock_client,
        mock_box_session,
        search_response,
        file_id,
        folder_id,
):
    # pylint:disable=redefined-outer-name
    mock_box_session.get.return_value = search_response
    search_term = 'lolcatz'
    search_result = mock_client.search(
        search_term,
        10,
        0,
        ancestor_folders=[Folder(mock_box_session, folder_id)],
        file_extensions=['.jpg'],
    )
    assert search_result[0].object_id == file_id
    assert search_result[1].object_id == folder_id


def test_events_returns_event_object(mock_client):
    # pylint:disable=redefined-outer-name
    assert isinstance(mock_client.events(), Events)


def test_groups_return_the_correct_group_objects(
        mock_client,
        mock_box_session,
        groups_response,
        group_id_1,
        group_id_2,
):
    # pylint:disable=redefined-outer-name
    mock_box_session.get.return_value = groups_response
    groups = mock_client.groups()
    for group, expected_id in zip(groups, [group_id_1, group_id_2]):
        assert group.object_id == expected_id
        assert group.name == str(expected_id)
        # pylint:disable=protected-access
        assert group._session == mock_box_session


def test_create_group_returns_the_correct_group_object(mock_client, mock_box_session, create_group_response):
    # pylint:disable=redefined-outer-name
    test_group_name = 'test_group_name'
    value = json.dumps({'name': test_group_name})
    mock_box_session.post.return_value = create_group_response
    new_group = mock_client.create_group(name=test_group_name)

    assert len(mock_box_session.post.call_args_list) == 1

    assert mock_box_session.post.call_args[0] == ("{0}/groups".format(API.BASE_API_URL),)
    assert mock_box_session.post.call_args[1] == {'data': value}
    assert isinstance(new_group, Group)
    assert new_group.object_id == 1234
    assert new_group.name == test_group_name


@pytest.mark.parametrize('password', (None, 'p4ssw0rd'))
def test_get_shared_item_returns_the_correct_item(mock_client, mock_box_session, shared_item_response, password):
    # pylint:disable=redefined-outer-name
    shared_link = 'https://cloud.box.com/s/661wcw2iz6q5r7v5xxkm'
    mock_box_session.request.return_value = shared_item_response
    item = mock_client.get_shared_item(shared_link, password)
    assert item.type == shared_item_response.json()['type']
    mock_box_session.request.assert_called_once_with(
        'GET',
        '{0}/shared_items'.format(API.BASE_API_URL),
        headers={
            'BoxApi': 'shared_link={0}{1}'.format(
                shared_link, '&shared_link_password={0}'.format(password) if password is not None else ''
            ),
        },
    )


@pytest.mark.parametrize('test_method', [
    'get',
    'post',
    'put',
    'delete',
    'options',
])
def test_make_request_passes_request_on_to_session(mock_client, mock_box_session, test_method):
    # pylint:disable=redefined-outer-name
    mock_client.make_request(test_method, 'url')
    assert mock_box_session.request.call_args[0] == (test_method, 'url')



# coding: utf-8

from __future__ import unicode_literals

from abc import ABCMeta, abstractmethod
from six import add_metaclass


@add_metaclass(ABCMeta)
class Network(object):
    """
    Abstract base class specifying the interface of the network layer.
    """

    @abstractmethod
    def request(self, method, url, access_token, **kwargs):
        """
        Make a network request to the given url with the given method.

        :param method:
            The HTTP verb that should be used to make the request.
        :type method:
            `unicode`
        :param url:
            The URL for the request.
        :type url:
            `unicode`
        :param access_token:
            The OAuth2 access token used to authorize the request.
        :type access_token:
            `unicode`
        """
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def retry_after(self, delay, request_method, *args, **kwargs):
        """
        Make a network request after a given delay.

        :param delay:
            How long until the request should be executed.
        :type delay:
            `float`
        :param request_method:
            A callable that will execute the request.
        :type request_method:
            `callable`
        """
        raise NotImplementedError  # pragma: no cover


@add_metaclass(ABCMeta)
class NetworkResponse(object):
    """Abstract base class specifying the interface for a network response."""

    @abstractmethod
    def json(self):
        """Return the parsed JSON response.

        :rtype:
            `dict` or `list` or `str` or `int` or `float`
        """
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def content(self):
        """Return the content of the response body.

        :rtype:
            varies
        """
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def status_code(self):
        """Return the HTTP status code of the response.

        :rtype:
            `int`
        """
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def ok(self):
        """Return whether or not the request was successful.

        :rtype:
            `bool`
        """
        # pylint:disable=invalid-name
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def headers(self):
        """Return the response headers.

        :rtype:
            `dict`
        """
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def response_as_stream(self):
        """Return a stream containing the raw network response.

        :rtype:
            `stream`
        """
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def access_token_used(self):
        """Return the access token used to make the request.

        :rtype:
            `unicode`
        """
        raise NotImplementedError  # pragma: no cover



# coding: utf-8

from __future__ import unicode_literals


class API(object):
    """Configuration object containing the URLs for the Box API."""
    BASE_API_URL = 'https://api.box.com/2.0'
    UPLOAD_URL = 'https://upload.box.com/api/2.0'
    OAUTH2_API_URL = 'https://api.box.com/oauth2'  # <https://developers.box.com/docs/#oauth-2>
    OAUTH2_AUTHORIZE_URL = 'https://app.box.com/api/oauth2/authorize'  # <https://developers.box.com/docs/#oauth-2-authorize>



# coding: utf-8

from __future__ import unicode_literals
from boxsdk.exception import BoxAPIException, BoxOAuthException


def test_box_api_exception():
    status = 'status'
    code = 400
    message = 'message'
    request_id = 12345
    headers = {'header': 'value'}
    url = 'https://example.com'
    method = 'GET'
    context_info = {'context': 'value'}
    box_exception = BoxAPIException(
        status,
        code=code,
        message=message,
        request_id=request_id,
        headers=headers,
        url=url,
        method=method,
        context_info=context_info,
    )
    assert box_exception.status == status
    assert box_exception.code == code
    assert box_exception.message == message
    assert box_exception.request_id == request_id
    assert box_exception._headers == headers  # pylint:disable=protected-access
    assert box_exception.url == url
    assert box_exception.method == method
    assert str(box_exception) == '''
Message: {0}
Status: {1}
Code: {2}
Request id: {3}
Headers: {4}
URL: {5}
Method: {6}
Context info: {7}'''.format(message, status, code, request_id, headers, url, method, context_info)


def test_box_oauth_exception():
    status = 'status'
    message = 'message'
    url = 'https://example.com'
    method = 'GET'
    box_exception = BoxOAuthException(
        status,
        message=message,
        url=url,
        method=method,
    )
    assert str(box_exception) == '''
Message: {0}
Status: {1}
URL: {2}
Method: {3}'''.format(message, status, url, method)



# coding: utf-8

from __future__ import unicode_literals

from boxsdk.config import API


def test_user_url(mock_user):
    # pylint:disable=redefined-outer-name, protected-access
    assert mock_user.get_url() == '{0}/{1}/{2}'.format(API.BASE_API_URL, 'users', mock_user.object_id)



# coding: utf-8

from __future__ import unicode_literals
import pytest
from boxsdk.exception import BoxAPIException
from test.functional.mock_box.util.chaos_utils import error, html, xml
from test.functional.mock_box.util.http_utils import RETRY_AFTER_HEADER


@pytest.fixture(params=[500, 501])
def error_code(request):
    return request.param


@pytest.fixture(params=[202, 429])
def retry_code(request):
    return request.param


@pytest.fixture(autouse=True)
def mock_sleep(monkeypatch):
    monkeypatch.setattr('time.sleep', lambda seconds: None)


@pytest.fixture(params=[1, 2, (1, 2), (1, 3)])
def should_apply(request):
    expected_num_requests = 4  # 3 auth requests and 1 folder info request
    # Figure out how many times the call should fail. If should_apply is an int, it should fail iff should_apply == 1
    # If it's a sequence, figure out the first time it will succeed. It should fail until then.
    if isinstance(request.param, int):
        if request.param is 1:
            expected_num_requests += 1
    else:
        expected_num_requests += next(
            (a for a in enumerate(sorted(request.param)) if a[0] + 1 != a[1]),
            [0],
        )[0] or len(request.param)
    return request.param, expected_num_requests


@pytest.fixture(autouse=True)
def reset_chaos(mock_box, request):
    mock_box.get_folder_info.reset_chaos()
    request.addfinalizer(mock_box.get_folder_info.reset_chaos)


def test_client_retries_on_server_error(box_client, mock_box, error_code, should_apply):
    # pylint:disable=redefined-outer-name
    should_apply, expected_num_requests = should_apply
    mock_box.get_folder_info.add_chaos(error(error_code), should_apply)
    box_client.folder(0).get()
    assert len(mock_box.requests) == expected_num_requests


def test_client_retries_on_retry_after(box_client, mock_box, retry_code, should_apply):
    # pylint:disable=redefined-outer-name
    should_apply, expected_num_requests = should_apply
    mock_box.get_folder_info.add_chaos(error(retry_code, headers={RETRY_AFTER_HEADER: 1}), should_apply)
    box_client.folder(0).get()
    assert len(mock_box.requests) == expected_num_requests


def test_client_stops_retrying_after_10_server_errors(box_client, mock_box, error_code):
    # pylint:disable=redefined-outer-name
    mock_box.get_folder_info.add_chaos(error(error_code))
    with pytest.raises(BoxAPIException) as exc_info:
        box_client.folder(0).get()
        assert exc_info.value.status == error_code
    assert len(mock_box.requests) == 14  # 3 auth requests, 1 try, and 10 retries


@pytest.mark.parametrize('chaos', [html, xml])
def test_non_json_response_raises(box_client, mock_box, chaos):
    # pylint:disable=redefined-outer-name
    mock_box.get_folder_info.add_chaos(chaos)
    with pytest.raises(BoxAPIException) as exc_info:
        box_client.folder(0).get()
        assert exc_info.value.status == 200
        assert 'json' in exc_info.value.message


def test_unknown_resource_raises(box_client):
    folder = box_client.folder('0')
    with pytest.raises(BoxAPIException) as exc_info:
        # pylint:disable=protected-access
        box_client.make_request('get', folder.get_url('unknown')).get_info()
        # pylint:enable=protected-access
        assert exc_info.value.status == 404


def test_make_request_equivalent_to_get_items(box_client):
    folder = box_client.folder('0')
    # pylint:disable=protected-access
    response = box_client.make_request('get', folder.get_url())
    # pylint:enable=protected-access
    assert response.status_code == 200



# coding: utf-8

from __future__ import unicode_literals
from threading import Event, Thread
import pytest
from six import add_metaclass
from boxsdk.util.singleton import Singleton

# pylint:disable=redefined-outer-name


@pytest.fixture
def singleton_class():
    @add_metaclass(Singleton)
    class MySingleton(object):
        def __init__(self, on_enter=None, block_on=None):
            if on_enter:
                on_enter.set()
            if block_on:
                block_on.wait()
    return MySingleton


@pytest.fixture
def nested_singleton(singleton_class):
    @add_metaclass(Singleton)
    class Nested(object):
        def __init__(self):
            self.single = singleton_class()
    return Nested


def test_singleton_returns_singleton(singleton_class):
    instance1 = singleton_class()
    instance2 = singleton_class()

    assert instance1 is instance2


def test_nested_singletons_are_safely_initialized(nested_singleton, singleton_class):
    """Ensure that the init of one Singleton can itself init another singleton."""
    nested = nested_singleton()
    single = singleton_class()

    assert nested.single is single


def test_initializing_singleton_is_atomic(singleton_class):
    """
    Forces a race condition on initialization and ensures that only 1 instances is created

    This test is setting up the race condition by creating 2 threads, each which will ask for a reference
    to the target singleton. The __init__ method of that singleton can be controlled by and communicate with
    this test code via interaction with 2 Event objects. The first thread will be launched and block
    (via Event.wait) inside the __init__ of the Singleton. And then the 2nd thread will be unleashed and the
    test code can ensure that this second thread can't get into the __init__, because of the atomicity of
    the singleton.

    Some my content that doing this test by mocking out the Lock is better/simpler. This implementation uses
    real threads and *zero* knowledge of the internals of the Singleton implementation. Thus it is less fragile.
    Singleton can change implementation to use some other Locking scheme... and nothing here changes. A mock version
    would break in that eventuality.
    """
    results = []
    thread1_inside_singleton, thread1_is_released = Event(), Event()
    thread2_inside_singleton, thread2_has_started = Event(), Event()

    def entry(on_enter, block_on, thread_has_started):
        if thread_has_started:
            thread_has_started.set()
        single = singleton_class(on_enter, block_on)
        results.append(single)

    thread1 = Thread(target=entry, args=(thread1_inside_singleton, thread1_is_released, None))
    thread1.start()

    thread1_inside_singleton.wait()

    thread2 = Thread(target=entry, args=(thread2_inside_singleton, None, thread2_has_started))
    thread2.start()

    thread2_has_started.wait()              # We block until the thread is known to have started.
    assert not thread2_inside_singleton.wait(0.01)  # Wait a bit ensuring that the thread never makes it inside
    assert thread2.is_alive()               # And yet the thread is still alive.

    thread1_is_released.set()               # releasing thread1 will allow both threads to complete.

    thread1.join()
    thread2.join()
    assert not thread2_inside_singleton.wait(0)   # Re-ensure that thread2 never made it inside the singleton __init__
    assert results[0] is results[1]         # And of course there's only 1 instance of the singleton



# coding: utf-8

from __future__ import unicode_literals
import json
from mock import Mock
import pytest
from requests import Response
from boxsdk import Client
from boxsdk.auth.oauth2 import OAuth2
from test.integration.mock_network import MockNetwork


@pytest.fixture()
def box_client(box_oauth, mock_box_network):
    # pylint:disable=redefined-outer-name
    return Client(box_oauth, network_layer=mock_box_network)


@pytest.fixture()
def box_oauth(mock_box_network, client_id, client_secret, access_token, refresh_token):
    # pylint:disable=redefined-outer-name
    return OAuth2(
        client_id,
        client_secret,
        network_layer=mock_box_network,
        access_token=access_token,
        refresh_token=refresh_token,
    )


@pytest.fixture(scope='module')
def mock_box_network():
    return MockNetwork()


@pytest.fixture(scope='session')
def generic_successful_response():
    mock_network_response = Mock(Response)
    content = '{"message": "success"}'
    mock_network_response.content = content.encode('utf-8')
    mock_network_response.status_code = 200
    mock_network_response.ok = True
    mock_network_response.raw = Mock()
    mock_network_response.json.return_value = json.loads(content)
    return mock_network_response


@pytest.fixture(scope='session')
def successful_token_mock():
    return Mock(Response)


@pytest.fixture(scope='session')
def unauthorized_response():
    res = Mock(Response)
    res.content = b''
    res.status_code = 401
    res.ok = False
    return res



# coding: utf-8

from __future__ import unicode_literals
from requests.packages.urllib3.exceptions import TimeoutError

from boxsdk.object.base_endpoint import BaseEndpoint
from boxsdk.util.lru_cache import LRUCache


class Events(BaseEndpoint):
    """Box API endpoint for subscribing to changes in a Box account."""

    def get_url(self, *args):
        """Base class override."""
        return super(Events, self).get_url('events', *args)

    def get_events(self, limit=100, stream_position=0, stream_type='all'):
        """
        Get Box events from a given stream position for a given stream type.

        :param limit:
            Maximum number of events to return.
        :type limit:
            `int`
        :param stream_position:
            The location in the stream from which to start getting events. 0 is the beginning of time. 'now' will
            return no events and just current stream position.
        :type stream_position:
            `unicode`
        :param stream_type:
            Which type of events to return. Can be 'all', 'tree', or 'sync'.
        :type stream_type:
            `unicode`
        :returns:
            JSON response from the Box /events endpoint. Contains the next stream position to use for the next call,
            along with some number of events.
        :rtype:
            `dict`
        """
        url = self.get_url()
        params = {
            'limit': limit,
            'stream_position': stream_position,
            'stream_type': stream_type,
        }
        box_response = self._session.get(url, params=params)
        return box_response.json()

    def get_latest_stream_position(self):
        """
        Get the latest stream position. The return value can be used with :meth:`get_events` or
        :meth:`generate_events_with_long_polling`.

        :returns:
            The latest stream position.
        :rtype:
            `unicode`
        """
        url = self.get_url()
        params = {
            'stream_position': 'now',
        }
        return self._session.get(url, params=params).json()['next_stream_position']

    def _get_all_events_since(self, stream_position):
        next_stream_position = stream_position
        while True:
            events = self.get_events(stream_position=next_stream_position, limit=100)
            next_stream_position = events['next_stream_position']
            events = events['entries']
            if not events:
                return
            for event in events:
                yield event, next_stream_position
            if len(events) < 100:
                return

    def long_poll(self, options, stream_position):
        """
        Set up a long poll connection at the specified url.

        :param options:
            The long poll options which include a long pull url, retry timeout, etc.
        :type options:
            `dict`
        :param stream_position:
            The location in the stream from which to start getting events. 0 is the beginning of time.
            'now' will return no events and just current stream position.
        :type stream_position:
            `unicode`
        :returns:
            {"message": "new_change"}, which means there're new changes on Box or {"version": 1, "message": "reconnect"}
            if nothing happens on Box during the long poll.
        :rtype:
            `dict`
        """
        url = options['url']
        long_poll_response = self._session.get(
            url,
            timeout=options['retry_timeout'],
            params={'stream_position': stream_position}
        )
        return long_poll_response

    def generate_events_with_long_polling(self, stream_position=None):
        """
        Subscribe to events from the given stream position.

        :param stream_position:
            The location in the stream from which to start getting events. 0 is the beginning of time. 'now' will
            return no events and just current stream position.
        :type stream_position:
            `unicode`
        :returns:
            Events corresponding to changes on Box in realtime, as they come in.
        :rtype:
            `generator` of :class:`Event`
        """
        event_ids = LRUCache()
        stream_position = stream_position if stream_position is not None else self.get_latest_stream_position()
        while True:
            options = self.get_long_poll_options()
            while True:
                try:
                    long_poll_response = self.long_poll(options, stream_position)
                except TimeoutError:
                    break
                else:
                    message = long_poll_response.json()['message']
                    if message == 'new_change':
                        next_stream_position = stream_position
                        for event, next_stream_position in self._get_all_events_since(stream_position):
                            try:
                                event_ids.get(event['event_id'])
                            except KeyError:
                                yield event
                                event_ids.set(event['event_id'])
                        stream_position = next_stream_position
                        break
                    elif message == 'reconnect':
                        continue
                    else:
                        break

    def get_long_poll_options(self):
        """
        Get the url and retry timeout for setting up a long polling connection.

        :returns:
            A `dict` including a long poll url, retry timeout, etc.
            E.g.
                {
                    "type": "realtime_server",
                    "url": "http://2.realtime.services.box.net/subscribe?channel=cc807c9c4869ffb1c81a&stream_type=all",
                    "ttl": "10",
                    "max_retries": "10",
                    "retry_timeout": 610,
                }
        :rtype:
            `dict`
        """
        url = self.get_url()
        box_response = self._session.options(url)
        return box_response.json()['entries'][0]



# coding: utf-8

from __future__ import unicode_literals
import os
from mock import Mock
import pytest
from six import int2byte, PY2
from boxsdk.object.collaboration import Collaboration
from boxsdk.object.file import File
from boxsdk.object.folder import Folder
from boxsdk.object.group import Group
from boxsdk.object.user import User
from boxsdk.object.search import Search


# pylint:disable=redefined-outer-name


@pytest.fixture(scope='module')
def mock_group_membership_id():
    return 'fake-group-membership-5'


@pytest.fixture(scope='module')
def mock_collaboration_id():
    return 'collab_id1'


@pytest.fixture(scope='module')
def mock_file_path():
    return os.path.join('path', 'to', 'file')


@pytest.fixture(scope='function')
def mock_content_response(make_mock_box_request):
    mock_box_response, mock_network_response = make_mock_box_request(content=b'Contents of a text file.')
    mock_network_response.response_as_stream = raw = Mock()
    raw.stream.return_value = (b if PY2 else int2byte(b) for b in mock_box_response.content)
    return mock_box_response


@pytest.fixture(scope='function')
def mock_upload_response(mock_object_id, make_mock_box_request):
    mock_box_response, _ = make_mock_box_request(
        response={'entries': [{'type': 'file', 'id': mock_object_id}]},
    )
    return mock_box_response


@pytest.fixture()
def test_collaboration(mock_box_session, mock_collaboration_id):
    return Collaboration(mock_box_session, mock_collaboration_id)


@pytest.fixture()
def test_file(mock_box_session, mock_object_id):
    return File(mock_box_session, mock_object_id)


@pytest.fixture()
def test_folder(mock_box_session, mock_object_id):
    return Folder(mock_box_session, mock_object_id)


@pytest.fixture()
def test_group(mock_box_session, mock_group_id):
    return Group(mock_box_session, mock_group_id)


@pytest.fixture()
def test_search(mock_box_session):
    return Search(mock_box_session)


@pytest.fixture(scope='function')
def mock_collab_response(make_mock_box_request, mock_collaboration_id):
    mock_box_response, _ = make_mock_box_request(
        response={'type': 'collaboration', 'id': mock_collaboration_id},
    )
    return mock_box_response


@pytest.fixture()
def mock_user(mock_box_session, mock_user_id):
    user = User(mock_box_session, mock_user_id)
    return user


@pytest.fixture(scope='function')
def mock_precondition_failed_response(mock_object_id, make_mock_box_request):
    mock_box_response, _ = make_mock_box_request(
        status_code=412,
        response_ok=False,
        response={'type': 'folder', 'id': mock_object_id},
    )
    return mock_box_response


@pytest.fixture()
def mock_group_membership_dict(mock_group_membership_id, mock_user_id, mock_group_id):
    data = {
        'type': 'group_membership',
        'id': mock_group_membership_id,
        'role': 'member',
        'user': {'type': 'user', 'id': mock_user_id},
        'group': {'type': 'group', 'id': mock_group_id},
    }
    return data


@pytest.fixture()
def mock_add_member_response(mock_group_membership_dict, make_mock_box_request):
    mock_box_response, _ = make_mock_box_request(
        status_code=200,
        response_ok=True,
        response=mock_group_membership_dict,
    )
    return mock_box_response


@pytest.fixture()
def if_match_header(etag):
    return {'If-Match': etag} if etag is not None else None


@pytest.fixture()
def if_none_match_header(etag):
    return {'If-None-Match': etag} if etag is not None else None


@pytest.fixture(params=[None, 'etag'])
def etag(request):
    return request.param


@pytest.fixture(params=[True, False])
def preflight_check(request):
    return request.param


@pytest.fixture(params=[True, False])
def upload_using_accelerator(request):
    return request.param


@pytest.fixture(params=[True, False])
def preflight_fails(preflight_check, request):
    return preflight_check and request.param


@pytest.fixture(params=[True, False])
def upload_using_accelerator_fails(upload_using_accelerator, request):
    return upload_using_accelerator and request.param


@pytest.fixture(params=[0, 100])
def file_size(request):
    return request.param


@pytest.fixture()
def mock_group(mock_box_session, mock_group_id):
    group = Group(mock_box_session, mock_group_id)
    return group



# coding: utf-8

from __future__ import unicode_literals
from functools import partial
from mock import Mock
import pytest
import re
from threading import Thread

from six.moves.urllib import parse as urlparse  # pylint:disable=import-error,no-name-in-module

from boxsdk.exception import BoxOAuthException
from boxsdk.network.default_network import DefaultNetworkResponse
from boxsdk.auth.oauth2 import OAuth2
from boxsdk.config import API


@pytest.fixture(params=('https://url.com/foo?bar=baz', 'https://ȕŕľ.com/ƒőő?Ƅȁŕ=Ƅȁż', None))
def redirect_url(request):
    """A value for the `redirect_uri` query string parameter for OAuth2."""
    return request.param


def test_get_correct_authorization_url(redirect_url):
    # pylint:disable=redefined-outer-name
    fake_client_id = 'fake_client_id'
    fake_client_secret = 'fake_client_secret'
    oauth2 = OAuth2(
        client_id=fake_client_id,
        client_secret=fake_client_secret,
    )
    auth_url, csrf_token = oauth2.get_authorization_url(redirect_url=redirect_url)
    expected_auth_url_format = '{0}?state={1}&response_type=code&client_id={2}'
    if redirect_url:
        expected_auth_url_format += '&redirect_uri={3}'
    assert auth_url == expected_auth_url_format.format(
        API.OAUTH2_AUTHORIZE_URL,
        csrf_token,
        fake_client_id,
        urlparse.quote_plus((redirect_url or '').encode('utf-8')),
    )
    assert re.match('^box_csrf_token_[A-Za-z0-9]{16}$', csrf_token)


def test_authenticate_send_post_request_with_correct_params(mock_network_layer, successful_token_response):
    fake_client_id = 'fake_client_id'
    fake_client_secret = 'fake_client_secret'
    fake_auth_code = 'fake_auth_code'
    data = {
        'grant_type': 'authorization_code',
        'code': fake_auth_code,
        'client_id': fake_client_id,
        'client_secret': fake_client_secret,
        'box_device_id': '0',
        'box_device_name': 'my_awesome_device',
    }
    mock_network_layer.request.return_value = successful_token_response
    oauth = OAuth2(
        client_id=fake_client_id,
        client_secret=fake_client_secret,
        network_layer=mock_network_layer,
        box_device_name='my_awesome_device',
    )

    oauth.authenticate(fake_auth_code)

    mock_network_layer.request.assert_called_once_with(
        'POST',
        '{0}/token'.format(API.OAUTH2_API_URL),
        data=data,
        headers={'content-type': 'application/x-www-form-urlencoded'},
        access_token=None,
    )

    assert oauth.access_token == successful_token_response.json()['access_token']


@pytest.mark.parametrize('_', range(10))
def test_refresh_send_post_request_with_correct_params_and_handles_multiple_requests(
        mock_network_layer,
        successful_token_response,
        _,
):
    fake_client_id = 'fake_client_id'
    fake_client_secret = 'fake_client_secret'
    fake_refresh_token = 'fake_refresh_token'
    fake_access_token = 'fake_access_token'
    data = {
        'grant_type': 'refresh_token',
        'refresh_token': fake_refresh_token,
        'client_id': fake_client_id,
        'client_secret': fake_client_secret,
        'box_device_id': '0',
        'box_device_name': 'my_awesome_device',
    }
    mock_network_layer.request.return_value = successful_token_response
    oauth = OAuth2(
        client_id=fake_client_id,
        client_secret=fake_client_secret,
        access_token=fake_access_token,
        refresh_token=fake_refresh_token,
        network_layer=mock_network_layer,
        box_device_name='my_awesome_device',
    )

    # Create four threads to call refresh on oauth at the same time.
    threads = []
    for _ in range(4):
        threads.append(Thread(target=oauth.refresh, args=(fake_access_token,)))

    for thread in threads:
        thread.start()
    for thread in threads:
        thread.join()

    # Assert that even four threads were trying to refresh the tokens at the same time, only one token request was made,
    # and it was made with the correct params.
    mock_network_layer.request.assert_called_once_with(
        'POST',
        '{0}/token'.format(API.OAUTH2_API_URL),
        data=data,
        headers={'content-type': 'application/x-www-form-urlencoded'},
        access_token=fake_access_token,
    )


def test_authenticate_stores_tokens_correctly(mock_network_layer, successful_token_response):
    fake_client_id = 'fake_client_id'
    fake_client_secret = 'fake_client_secret'
    fake_auth_code = 'fake_auth_code'

    mock_network_layer.request.return_value = successful_token_response
    mock_token_callback = Mock()
    oauth = OAuth2(
        client_id=fake_client_id,
        client_secret=fake_client_secret,
        network_layer=mock_network_layer,
        store_tokens=mock_token_callback,
    )

    access_token, refresh_token = oauth.authenticate(fake_auth_code)
    mock_token_callback.assert_called_once_with(access_token, refresh_token)

    assert access_token == successful_token_response.json()['access_token']
    assert refresh_token == successful_token_response.json()['refresh_token']


@pytest.mark.parametrize('_', range(10))
def test_refresh_gives_back_the_correct_response_and_handles_multiple_requests(
        mock_network_layer,
        successful_token_response,
        network_response_with_missing_tokens,
        _,
):
    # pylint:disable=redefined-outer-name
    fake_client_id = 'fake_client_id'
    fake_client_secret = 'fake_client_secret'
    fake_refresh_token = 'fake_refresh_token'
    fake_access_token = 'fake_access_token'

    # Setup the network layer so that if oauth makes more than one request, it will get a malformed response and failed
    # the test.
    mock_network_layer.request.side_effect = [successful_token_response, network_response_with_missing_tokens]
    oauth = OAuth2(
        client_id=fake_client_id,
        client_secret=fake_client_secret,
        access_token=fake_access_token,
        refresh_token=fake_refresh_token,
        network_layer=mock_network_layer,
    )

    def refresh_tokens_and_verify_the_response():
        access_token, refresh_token = oauth.refresh(fake_access_token)
        assert access_token == successful_token_response.json()['access_token']
        assert refresh_token == successful_token_response.json()['refresh_token']

    # Creates four threads and do token refresh at the same time. Assert they all get the same new access token and
    # refresh token.
    threads = []
    for _ in range(4):
        threads.append(Thread(target=refresh_tokens_and_verify_the_response))

    for thread in threads:
        thread.start()
    for thread in threads:
        thread.join()


@pytest.mark.parametrize('test_method', [
    partial(OAuth2.refresh, access_token_to_refresh='fake_access_token'),
    partial(OAuth2.authenticate, auth_code='fake_code')
])
def test_token_request_raises_box_oauth_exception_when_getting_bad_network_response(
        test_method,
        mock_network_layer,
        bad_network_response,
):
    with pytest.raises(BoxOAuthException):
        mock_network_layer.request.return_value = bad_network_response
        oauth = OAuth2(
            client_id='',
            client_secret='',
            access_token='fake_access_token',
            network_layer=mock_network_layer,
        )
        test_method(oauth)


@pytest.mark.parametrize('test_method', [
    partial(OAuth2.refresh, access_token_to_refresh='fake_access_token'),
    partial(OAuth2.authenticate, auth_code='fake_code')
])
def test_token_request_raises_box_oauth_exception_when_no_json_object_can_be_decoded(
        test_method,
        mock_network_layer,
        non_json_response,
):
    mock_network_layer.request.return_value = non_json_response
    oauth = OAuth2(
        client_id='',
        client_secret='',
        access_token='fake_access_token',
        network_layer=mock_network_layer,
    )
    with pytest.raises(BoxOAuthException):
        test_method(oauth)


@pytest.fixture(params=[
    ['access_token'],
    ['refresh_token'],
    [],
])
def network_response_with_missing_tokens(request):
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.ok = True
    json_dict = {}
    for key in request.param:
        json_dict[key] = 'fake_token'
    mock_network_response.json.return_value = json_dict
    return mock_network_response


@pytest.mark.parametrize('test_method', [
    partial(OAuth2.refresh, access_token_to_refresh='fake_access_token'),
    partial(OAuth2.authenticate, auth_code='fake_code')
])
def test_token_request_raises_box_oauth_exception_when_tokens_are_not_in_the_response(
        test_method,
        mock_network_layer,
        network_response_with_missing_tokens,
):
    # pylint:disable=redefined-outer-name
    mock_network_layer.request.return_value = network_response_with_missing_tokens
    oauth = OAuth2(
        client_id='',
        client_secret='',
        access_token='fake_access_token',
        network_layer=mock_network_layer,
    )
    with pytest.raises(BoxOAuthException):
        test_method(oauth)



# coding: utf-8

from __future__ import unicode_literals
from sqlalchemy.orm.exc import NoResultFound
from test.functional.mock_box.db_model.user_model import UserModel
from test.functional.mock_box.util.db_utils import get_user_from_header
from test.functional.mock_box.util.http_utils import abort


class UserBehavior(object):
    def __init__(self, db_session):
        self._db_session = db_session

    def _get_user_by_id(self, user_id):
        try:
            return self._db_session.query(UserModel).filter_by(user_id=user_id).one()
        except NoResultFound:
            abort(401)

    def get_user_info(self, user_id):
        if user_id == 'me':
            user_id = get_user_from_header(self._db_session).user_id
        else:
            user_id = int(user_id)
        user = self._get_user_by_id(user_id)
        return {
            'type': 'user',
            'login': user.login,
            'enterprise': None,
            'id': user.user_id,
            'name': user.name,
        }



# coding: utf-8

from __future__ import unicode_literals
from six import add_metaclass
from boxsdk.util.singleton import Singleton


@add_metaclass(Singleton)
class Translator(object):
    """
    Translate item responses from the Box API to Box objects.
    """
    def __init__(self):
        self._type_to_class_mapping = {}

    def register(self, type_name, box_cls):
        """
        Associate a Box object class to handle Box API item responses with the given type name.

        :param type_name:
            The type name to be registered.
        :type type_name:
            `unicode`
        :param box_cls:
            The Box object class, which will be associated with the type name provided.
        :type box_cls:
            `type`
        """
        self._type_to_class_mapping.update({
            type_name: box_cls,
        })

    def translate(self, type_name):
        """
        Get the box object class associated with the given type name.

        :param type_name:
            The type name to be translated.
        :type type_name:
            `unicode`
        """
        from boxsdk.object.base_object import BaseObject
        return self._type_to_class_mapping.get(type_name, BaseObject)



# coding: utf-8

from __future__ import unicode_literals
from mock import Mock
import requests
from boxsdk.network.default_network import DefaultNetworkResponse
from boxsdk.network.network_interface import Network


class MockNetwork(Network):
    """Mock implementation of the network interface for testing purposes."""

    def __init__(self):
        super(MockNetwork, self).__init__()
        self._session = Mock(requests.Session)
        self._retries = []

    def request(self, method, url, access_token, **kwargs):
        """Base class override.
        Make a mock network request using a mock requests.Session.
        """
        return DefaultNetworkResponse(self._session.request(method, url, **kwargs), access_token)

    def retry_after(self, delay, request_method, *args, **kwargs):
        """Base class override.
        Retry immediately, recording the retry request.
        """
        self._retries.append((delay, request_method, args, kwargs))
        return request_method(*args, **kwargs)

    @property
    def session(self):
        return self._session



# coding: utf-8

from __future__ import unicode_literals

import json

from .base_object import BaseObject
from boxsdk.config import API
from boxsdk.exception import BoxAPIException


class Item(BaseObject):
    """Box API endpoint for interacting with files and folders."""

    def _get_accelerator_upload_url(self, file_id=None):
        """
        Make an API call to get the Accelerator upload url for either upload a new file or updating an existing file.

        :param file_id:
            Box id of the file to be uploaded. Not required for new file uploads.
        :type file_id:
            `unicode` or None
        :return:
            The Accelerator upload url or None if cannot get the Accelerator upload url.
        :rtype:
            `unicode` or None
        """
        endpoint = '{0}/content'.format(file_id) if file_id else 'content'
        url = '{0}/files/{1}'.format(API.BASE_API_URL, endpoint)
        try:
            response_json = self._session.options(
                url=url,
                expect_json_response=True,
            ).json()
            return response_json.get('upload_url', None)
        except BoxAPIException:
            return None

    def _preflight_check(self, size, name=None, file_id=None, parent_id=None):
        """
        Make an API call to check if certain file can be uploaded to Box or not.
        (https://developers.box.com/docs/#files-preflight-check)

        :param size:
            The size of the file to be uploaded in bytes. Specify 0 for unknown file sizes.
        :type size:
            `int`
        :param name:
            The name of the file to be uploaded. This is optional if `file_id` is specified,
            but required for new file uploads.
        :type name:
            `unicode`
        :param file_id:
            Box id of the file to be uploaded. Not required for new file uploads.
        :type file_id:
            `unicode`
        :param parent_id:
            The ID of the parent folder. Required only for new file uploads.
        :type parent_id:
            `unicode`
        :raises:
            :class:`BoxAPIException` when preflight check fails.
        """
        endpoint = '{0}/content'.format(file_id) if file_id else 'content'
        url = '{0}/files/{1}'.format(API.BASE_API_URL, endpoint)
        data = {'size': size}
        if name:
            data['name'] = name
        if parent_id:
            data['parent'] = {'id': parent_id}

        self._session.options(
            url=url,
            expect_json_response=False,
            data=json.dumps(data),
        )

    def update_info(self, data, etag=None):
        """Baseclass override.

        :param etag:
            If specified, instruct the Box API to perform the update only if
            the current version's etag matches.
        :type etag:
            `unicode` or None
        :return:
            The updated object.
            Return a new object of the same type, without modifying the original object passed as self.
            Construct the new object with all the default attributes that are returned from the endpoint.
        :rtype:
            :class:`BaseObject`
        """
        # pylint:disable=arguments-differ
        headers = {'If-Match': etag} if etag is not None else None
        return super(Item, self).update_info(data, headers=headers)

    def rename(self, name):
        """
        Rename the item to a new name.

        :param name:
            The new name, you want the item to be renamed to.
        :type name:
            `unicode`
        """
        data = {
            'name': name,
        }
        return self.update_info(data)

    def get(self, fields=None, etag=None):
        """Base class override.

        :param etag:
            If specified, instruct the Box API to get the info only if the current version's etag doesn't match.
        :type etag:
            `unicode` or None
        :returns:
            Information about the file or folder.
        :rtype:
            `dict`
        :raises: :class:`BoxAPIException` if the specified etag matches the latest version of the item.
        """
        # pylint:disable=arguments-differ
        headers = {'If-None-Match': etag} if etag is not None else None
        return super(Item, self).get(fields=fields, headers=headers)

    def copy(self, parent_folder):
        """Copy the item to the given folder.

        :param parent_folder:
            The folder to which the item should be copied.
        :type parent_folder:
            :class:`Folder`
        """
        url = self.get_url('copy')
        data = {
            'parent': {'id': parent_folder.object_id}
        }
        box_response = self._session.post(url, data=json.dumps(data))
        response = box_response.json()
        return self.__class__(
            session=self._session,
            object_id=response['id'],
            response_object=response,
        )

    def move(self, parent_folder):
        """
        Move the item to the given folder.

        :param parent_folder:
            The parent `Folder` object, where the item will be moved to.
        :type parent_folder:
            `Folder`
        """
        data = {
            'parent': {'id': parent_folder.object_id}
        }
        return self.update_info(data)

    def get_shared_link(self, access=None, etag=None):
        """Get a shared link for the item with the given access permissions.

        :param access:
            Determines who can access the shared link. May be open, company, or collaborators. If no access is
            specified, the default access will be used.
        :type access:
            `unicode` or None
        :param etag:
            If specified, instruct the Box API to create the link only if the current version's etag matches.
        :type etag:
            `unicode` or None
        :returns:
            The URL of the shared link.
        :rtype:
            `unicode`
        :raises: :class:`BoxAPIException` if the specified etag doesn't match the latest version of the item.
        """
        data = {
            'shared_link': {} if not access else {
                'access': access
            }
        }
        item = self.update_info(data, etag=etag)
        return item.shared_link['url']

    def remove_shared_link(self, etag=None):
        """Delete the shared link for the item.

        :param etag:
            If specified, instruct the Box API to delete the link only if the current version's etag matches.
        :type etag:
            `unicode` or None
        :returns:
            Whether or not the update was successful.
        :rtype:
            `bool`
        :raises: :class:`BoxAPIException` if the specified etag doesn't match the latest version of the item.
        """
        data = {'shared_link': None}
        item = self.update_info(data, etag=etag)
        return item.shared_link is None

    def delete(self, params=None, etag=None):
        """Delete the item.

        :param params:
            Additional parameters to send with the request.
        :type params:
            `dict`
        :param etag:
            If specified, instruct the Box API to delete the item only if the current version's etag matches.
        :type etag:
            `unicode` or None
        :returns:
            Whether or not the delete was successful.
        :rtype:
            `bool`
        :raises: :class:`BoxAPIException` if the specified etag doesn't match the latest version of the item.
        """
        headers = {'If-Match': etag} if etag is not None else None
        return super(Item, self).delete(params, headers)



# coding: utf-8

from __future__ import unicode_literals
import pytest
from boxsdk.exception import BoxOAuthException


def test_expired_access_token_is_refreshed(box_oauth, box_client, mock_box):
    # pylint:disable=protected-access
    mock_box.oauth.expire_token(box_oauth._access_token)
    # pylint:enable=protected-access
    box_client.folder(0).get()
    assert len(mock_box.requests) == 6  # GET /authorize, POST /authorize, /token, get_info, refresh /token, get_info


def test_expired_refresh_token_raises(box_oauth, box_client, mock_box):
    # pylint:disable=protected-access
    mock_box.oauth.expire_token(box_oauth._access_token)
    mock_box.oauth.expire_token(box_oauth._refresh_token)
    # pylint:enable=protected-access
    with pytest.raises(BoxOAuthException):
        box_client.folder(0).get()



# coding: utf-8

from __future__ import unicode_literals
from threading import RLock


class Singleton(type):
    """
    Metaclass for implementing the singleton pattern.

    Sample usage:
        @add_metaclass(Singleton)
        class my_singleton(object):
            def __init__(self):
                pass
    """
    _instances = {}
    _lock = RLock()

    def __call__(cls, *args, **kwargs):
        """
        When the class is instantiated, return the singleton if it exists; else create it and store for the next use.
        """
        if cls not in Singleton._instances:
            # Using an RLock, in case the singleton being initialized itself initializes another singleton.
            # RLock's are slow... but the 'if' just above ensures that slowness is only suffered when any
            # Singleton subclass is first initialized. After that the outer 'if' above will skip this code.
            with Singleton._lock:
                if cls not in Singleton._instances:
                    Singleton._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return Singleton._instances[cls]



# coding: utf-8

from __future__ import unicode_literals
import pytest
from boxsdk.object.base_object import BaseObject
from boxsdk.object.file import File
from boxsdk.object.folder import Folder
from boxsdk.object.group import Group
from boxsdk.object.user import User
from boxsdk.util.translator import Translator


_response_to_class_mapping = {}


@pytest.fixture
def bookmark_response(make_mock_box_request, mock_object_id):
    # pylint:disable=redefined-outer-name
    mock_box_response, _ = make_mock_box_request(
        response={'type': 'bookmark', 'id': mock_object_id},
    )
    return mock_box_response


@pytest.fixture
def box_note_response(make_mock_box_request, mock_object_id):
    # pylint:disable=redefined-outer-name
    mock_box_response, _ = make_mock_box_request(
        response={'type': 'boxnote', 'id': mock_object_id},
    )
    return mock_box_response


@pytest.fixture(autouse=True)
def translator_response(
        bookmark_response,
        box_note_response,
        mock_file_response,
        mock_folder_response,
        mock_group_response,
        mock_user_response,
):
    # pylint:disable=redefined-outer-name
    _response_to_class_mapping['bookmark'] = (bookmark_response, BaseObject)
    _response_to_class_mapping['box_note'] = (box_note_response, BaseObject)
    _response_to_class_mapping['file'] = (mock_file_response, File)
    _response_to_class_mapping['folder'] = (mock_folder_response, Folder)
    _response_to_class_mapping['group'] = (mock_group_response, Group)
    _response_to_class_mapping['user'] = (mock_user_response, User)


@pytest.mark.parametrize('response_type', ['bookmark', 'box_note', 'file', 'folder', 'group', 'user'])
def test_translator_converts_response_to_correct_type(response_type):
    response, object_class = _response_to_class_mapping[response_type]
    assert type(Translator().translate(response.json()['type']) == object_class)



# coding: utf-8

from __future__ import unicode_literals
from mock import ANY, call
from boxsdk.config import API


def test_automatic_refresh(
        box_client,
        mock_box_network,
        generic_successful_response,
        successful_token_response,
        unauthorized_response,
):
    mock_box_network.session.request.side_effect = [
        unauthorized_response,
        successful_token_response,
        generic_successful_response,
    ]
    box_client.folder(0).get()
    assert mock_box_network.session.request.mock_calls == [
        call(
            'GET',
            '{0}/folders/0'.format(API.BASE_API_URL),
            headers=ANY,
            params=None,
        ),
        call(
            'POST',
            '{0}/token'.format(API.OAUTH2_API_URL),
            data=ANY,
            headers={'content-type': 'application/x-www-form-urlencoded'},
        ),
        call(
            'GET',
            '{0}/folders/0'.format(API.BASE_API_URL),
            headers=ANY,
            params=None,
        ),
    ]



# coding: utf-8

from __future__ import unicode_literals
import json
from mock import Mock
import pytest
from boxsdk.auth.oauth2 import DefaultNetwork
from boxsdk.network import default_network
from boxsdk.network.default_network import DefaultNetworkResponse
from boxsdk.session.box_session import BoxResponse, BoxSession


@pytest.fixture()
def mock_box_session():
    return Mock(BoxSession)


@pytest.fixture(scope='function')
def mock_network_layer():
    mock_network = Mock(DefaultNetwork)
    return mock_network


@pytest.fixture(autouse=True)
def prevent_tests_from_making_real_network_requests(monkeypatch):
    monkeypatch.delattr(default_network.requests.Session, 'request')


@pytest.fixture(scope='function')
def mock_user_response(mock_user_id, make_mock_box_request):
    # pylint:disable=redefined-outer-name
    mock_box_response, _ = make_mock_box_request(
        response={'type': 'user', 'id': mock_user_id},
    )
    return mock_box_response


@pytest.fixture(scope='function')
def mock_group_response(mock_group_id, make_mock_box_request):
    # pylint:disable=redefined-outer-name
    mock_box_response, _ = make_mock_box_request(
        response={'type': 'group', 'id': mock_group_id},
    )
    return mock_box_response


@pytest.fixture(scope='module')
def mock_object_id():
    return '42'


@pytest.fixture(scope='module')
def mock_user_id():
    return 'fake-user-100'


@pytest.fixture(scope='module')
def mock_group_id():
    return 'fake-group-99'


@pytest.fixture()
def make_mock_box_request():
    def inner(status_code=200, response_ok=True, response=None, content=None):
        mock_box_response = Mock(BoxResponse)
        mock_network_response = Mock(DefaultNetworkResponse)
        mock_box_response.network_response = mock_network_response
        mock_box_response.status_code = status_code
        mock_box_response.ok = response_ok
        if response is not None:
            mock_box_response.json.return_value = response
            mock_box_response.content = json.dumps(response).encode()
        else:
            mock_box_response.content = content
        return mock_box_response, mock_network_response
    return inner


@pytest.fixture(scope='function')
def mock_file_response(mock_object_id, make_mock_box_request):
    # pylint:disable=redefined-outer-name
    mock_box_response, _ = make_mock_box_request(
        response={'type': 'file', 'id': mock_object_id},
    )
    return mock_box_response


@pytest.fixture(scope='function')
def mock_folder_response(mock_object_id, make_mock_box_request):
    # pylint:disable=redefined-outer-name
    mock_box_response, _ = make_mock_box_request(
        response={'type': 'folder', 'id': mock_object_id},
    )
    return mock_box_response



# coding: utf-8

from __future__ import unicode_literals
from itertools import product
import json
import pytest
from boxsdk.config import API
from boxsdk.object.base_object import BaseObject


@pytest.fixture(params=('file', 'folder', 'user'))
def test_object_and_response(
        test_file, test_folder, mock_user,
        mock_file_response, mock_folder_response, mock_user_response,
        request):
    test_objects_and_responses = {
        'file': (test_file, mock_file_response),
        'folder': (test_folder, mock_folder_response),
        'user': (mock_user, mock_user_response),
    }
    return test_objects_and_responses[request.param]


@pytest.mark.parametrize('params,headers', product(*([[None, {}, {'foo': 'bar'}, {'foo': 'bar', 'num': 4}]] * 2)))
def test_update_info(test_object_and_response, mock_box_session, params, headers):
    # pylint:disable=redefined-outer-name, protected-access
    test_object, mock_object_response = test_object_and_response
    expected_url = test_object.get_url()
    mock_box_session.put.return_value = mock_object_response
    data = {'foo': 'bar', 'baz': {'foo': 'bar'}, 'num': 4}
    update_response = BaseObject.update_info(test_object, data, params=params, headers=headers)
    mock_box_session.put.assert_called_once_with(expected_url, data=json.dumps(data), params=params, headers=headers)
    assert isinstance(update_response, test_object.__class__)
    assert update_response.object_id == test_object.object_id


@pytest.mark.parametrize('params, headers, success', [
    (None, None, True),
    ({'a': 'b'}, {'10': '20'}, True),
    ({'a': 'b'}, None, False),
])
def test_delete_handles_params_and_headers_correctly(mock_box_session, make_mock_box_request, params, headers, success):
    # pylint:disable=redefined-outer-name, protected-access
    fake_id = 'a_fake_id'
    base_object = BaseObject(mock_box_session, fake_id)

    mock_box_response, _ = make_mock_box_request(response_ok=success)
    mock_box_session.delete.return_value = mock_box_response
    expected_url = '{0}/{1}s/{2}'.format(API.BASE_API_URL, None, fake_id)
    update_response = base_object.delete(params=params, headers=headers)
    mock_box_session.delete.assert_called_once_with(
        expected_url,
        expect_json_response=False,
        params=params or {},
        headers=headers,
    )
    assert update_response is success


def test_getattr_and_getitem(test_object_and_response, mock_box_session):
    # pylint:disable=redefined-outer-name, protected-access
    test_object, mock_object_response = test_object_and_response
    mock_box_session.put.return_value = mock_object_response
    update_response = BaseObject.update_info(test_object, {})
    assert isinstance(update_response, test_object.__class__)
    assert update_response.object_id == update_response.id == update_response['id']


def test_get_url(test_object_and_response):
    # pylint:disable=redefined-outer-name, protected-access
    test_object, _ = test_object_and_response
    url = test_object.get_url()
    assert '{0}'.format(test_object.object_id) in url
    assert '{0}'.format(test_object._item_type) in url


def test_get_type_url(test_object_and_response):
    # pylint:disable=redefined-outer-name, protected-access
    test_object, _ = test_object_and_response
    url = test_object.get_type_url()
    assert url.endswith('{0}s'.format(test_object._item_type))



# coding: utf-8

from __future__ import unicode_literals

from .base_endpoint import BaseEndpoint
from boxsdk.util.translator import Translator
import json


class MetadataSearchFilter(object):
    """
    Helper class to encapsulate a single search filter. A search filter can only search against one template,
    but can filter on many fields.
    See :class:`MetadataSearchFilters`.
    """
    def __init__(self, template_key, scope):
        """
        :param template_key:
            The key of the template to search on
        :type template_key:
            `unicode`
        :param scope:
            The scope of the template to search on
        :type scope:
            `unicode`
        """
        self._template_key = template_key
        self._scope = scope
        self._field_filters = {}

    def as_dict(self):
        """
        Returns a `dict` representation of this object

        :return:
            The `dict` representation
        :rtype:
            `dict`
        """
        return {
            'templateKey': self._template_key,
            'scope': self._scope,
            'filters': self._field_filters
        }

    def add_value_based_filter(self, field_key, value):
        """
        Add a value-based filter (used for token-based search on string fields, and exact match search on all other fields)

        :param field_key:
            The field key to filter on
        :type field_filter:
            `unicode`
        :param value:
            The value to use to filter
        :type value:
            `unicode`
        """
        self._field_filters.update({field_key: value})

    def add_range_filter(self, field_key, gt_value=None, lt_value=None):
        """
        Add a range filter (used for ranged searches on numbers and dates)

        :param field_key:
            The field key to filter on
        :type field_filter:
            `unicode`
        :param gt_value:
            The lower bound of the range filter (inclusive)
        :type gt_value:
            `unicode` or `int` or `float` or `long` or None
        :param lt_value:
            The upper bound of the range filter (inclusive)
        :type lt_value:
            `unicode` or `int` or `float` or `long` or None
        """
        range_part = {}
        if gt_value:
            range_part['gt'] = gt_value
        if lt_value:
            range_part['lt'] = lt_value
        if not range_part:
            raise ValueError('Should specify gt and/or lt')
        self._field_filters.update({field_key: range_part})


class MetadataSearchFilters(object):
    """
    Helper class to encapsulate a list of metadata search filter params (mdfilters API param)
    See https://developers.box.com/metadata-api/#search for more details
    """
    def __init__(self):
        self._filters = []

    def as_list(self):
        """
        Get a list of filters from this object to use as a parameter in the Search API

        :return:
            The list of filters
        :rtype:
            `list` of `dict`
        """
        return [metadata_filter.as_dict() for metadata_filter in self._filters]

    def add_filter(self, metadata_filter):
        """
        Add a filter to this object. Note that the API only supports one filter.

        :param metadata_filter:
            The filter to add
        :type metadata_filter:
            :class:`MetadataSearchFilter`
        """
        self._filters.append(metadata_filter)


class Search(BaseEndpoint):
    """Search Box for files and folders."""

    def get_url(self, *args):
        """
        Gets the search endpoint URL.

        :return:
            The search endpoint URL.
        :rtype:
            `unicode`
        """
        return super(Search, self).get_url('search')

    @staticmethod
    def start_metadata_filters():
        """
        Get a :class:`MetadataSearchFilters` that represents a set of metadata filters.

        :return:
            The new :class:`MetadataSearchFilters`
        :rtype:
            :class:`MetadataSearchFilters`
        """
        return MetadataSearchFilters()

    @staticmethod
    def make_single_metadata_filter(template_key, scope):
        """
        Make a single :class:`MetadataSearchFilter` that represents a filter on a template. It must be
        added to a :class:`MetadataSearchFilters`.

        :return:
            The new :class:`MetadataSearchFilter`
        :rtype:
            :class:`MetadataSearchFilter`
        """
        return MetadataSearchFilter(template_key, scope)

    def search(
            self,
            query,
            limit=100,
            offset=0,
            ancestor_folders=None,
            file_extensions=None,
            metadata_filters=None,
            result_type=None,
            content_types=None,
            **kwargs
    ):
        """
        Search Box for items matching the given query.

        :param query:
            The string to search for.
        :type query:
            `unicode`
        :param limit:
            The maximum number of items to return.
        :type limit:
            `int`
        :param offset:
            The search result at which to start the response.
        :type offset:
            `int`
        :param ancestor_folders:
            Folder ids to limit the search to.
        :type ancestor_folders:
            `Iterable` of :class:`Folder`
        :param file_extensions:
            File extensions to limit the search to.
        :type file_extensions:
            `iterable` of `unicode`
        :param metadata_filters:
            Filters used for metadata search
        :type metadata_filters:
            :class:`MetadataSearchFilters`
        :param result_type:
            Which type of result you want. Can be file or folder.
        :type result_type:
            `unicode`
        :param content_types:
            Which content types to search. Valid types include name, description, file_content, comments, and tags.
        :type content_types:
            `Iterable` of `unicode`
        :return:
            A list of items that match the search query.
        :rtype:
            `list` of :class:`Item`
        """
        url = self.get_url()
        params = {
            'query': query,
            'limit': limit,
            'offset': offset,
        }
        if ancestor_folders:
            params.update({
                'ancestor_folder_ids': ','.join([folder.object_id for folder in ancestor_folders])
            })
        if file_extensions:
            params.update({'file_extensions': ','.join(file_extensions)})
        if metadata_filters:
            params.update({'mdfilters': json.dumps(metadata_filters.as_list())})
        if content_types:
            params.update({'content_types': ','.join(content_types)})
        if result_type:
            params.update({'type': result_type})
        params.update(kwargs)
        box_response = self._session.get(url, params=params)
        response = box_response.json()
        return [Translator().translate(item['type'])(self._session, item['id'], item) for item in response['entries']]



# coding: utf-8

from __future__ import unicode_literals

from boxsdk.exception import BoxAPIException


class BoxResponse(object):
    """Represents a response to a Box API request."""

    def __init__(self, network_response):
        self._network_response = network_response

    def json(self):
        """Return the parsed JSON response.

        :rtype:
            `dict` or `list` or `str` or `int` or `float`
        """
        return self._network_response.json()

    @property
    def content(self):
        """Return the content of the response body.

        :rtype:
            varies
        """
        return self._network_response.content

    @property
    def ok(self):
        """Return whether or not the request was successful.

        :rtype:
            `bool`
        """
        # pylint:disable=invalid-name
        return self._network_response.ok

    @property
    def status_code(self):
        """Return the HTTP status code of the response.

        :rtype:
            `int`
        """
        return self._network_response.status_code

    @property
    def network_response(self):
        """Return the underlying network response.

        :rtype:
            :class:`NetworkResponse`
        """
        return self._network_response


class BoxSession(object):
    """
    Box API session. Provides auth, automatic retry of failed requests, and session renewal.
    """

    def __init__(self, oauth, network_layer):
        """
        :param oauth:
            OAuth2 object used by the session to authorize requests.
        :type oauth:
            :class:`OAuth2`
        :param network_layer:
            Network implementation used by the session to make requests.
        :type network_layer:
            :class:`Network`
        """
        self._oauth = oauth
        self._network_layer = network_layer

    def _renew_session(self, access_token_used):
        """
        Renews the session by refreshing the access token.

        :param access_token_used:
            The access token that's currently being used by the session, that needs to be refreshed.
        :type access_token_used:
            `unicode`
        """
        self._oauth.refresh(access_token_used)

    @staticmethod
    def _is_json_response(network_response):
        """Return whether or not the network response content is json.

        :param network_response:
            The response from the Box API.
        :type network_response:
            :class:`NetworkResponse`
        """
        try:
            network_response.json()
            return True
        except ValueError:
            return False

    def _retry_request_if_necessary(self, network_response, attempt_number, *args, **kwargs):
        """
        Retry a request for certain types of failure.
        For 401 Unauthorized responses, renew the session by refreshing the access token; then retry.
        For 202 Accepted (thumbnail or file not ready) and 429 (too many requests), retry later, after a delay
        specified by the Retry-After header.
        For 5xx Server Error, retry later, after a delay; use exponential backoff to determine the delay.

        :param network_response:
            The response from the Box API.
        :type network_response:
            :class:`NetworkResponse`
        :param attempt_number:
            How many attempts at this request have already been tried. Used for exponential backoff calculations.
        :type attempt_number:
            `int`
        """
        if network_response.status_code == 401 and kwargs['auto_session_renewal']:
            self._renew_session(network_response.access_token_used)
            kwargs['auto_session_renewal'] = False
            return self._make_request(*args, **kwargs)
        elif network_response.status_code == 202 or network_response.status_code == 429:
            return self._network_layer.retry_after(
                float(network_response.headers['Retry-After']),
                self._make_request,
                *args,
                **kwargs
            )
        elif network_response.status_code >= 500 and attempt_number < 10:
            return self._network_layer.retry_after(
                2 ** attempt_number,
                self._make_request,
                *args,
                attempt_number=attempt_number + 1,
                **kwargs
            )
        return network_response

    def _raise_on_unsuccessful_request(self, network_response, expect_json_response, method, url):
        """
        Raise an exception if the request was unsuccessful.

        :param network_response:
            The network response which is being tested for success.
        :type network_response:
            :class:`NetworkResponse`
        :param expect_json_response:
            Whether or not the response content should be json.
        :type expect_json_response:
            `bool`
        :param method:
            The HTTP verb used to make the request.
        :type method:
            `unicode`
        :param url:
            The request URL.
        :type url:
            `unicode`
        """
        if not network_response.ok:
            response_json = {}
            try:
                response_json = network_response.json()
            except ValueError:
                pass
            raise BoxAPIException(
                status=network_response.status_code,
                headers=network_response.headers,
                code=response_json.get('code', None),
                message=response_json.get('message', None),
                request_id=response_json.get('request_id', None),
                url=url,
                method=method,
                context_info=response_json.get('context_info', None),
            )
        if expect_json_response and not self._is_json_response(network_response):
            raise BoxAPIException(
                status=network_response.status_code,
                headers=network_response.headers,
                message='Non-json response received, while expecting json response.',
                url=url,
                method=method,
            )

    def _prepare_and_send_request(
            self,
            method,
            url,
            headers=None,
            auto_session_renewal=True,
            expect_json_response=True,
            attempt_number=0,
            **kwargs
    ):
        """
        Prepare a request to be sent to the Box API.

        :param method:
            The HTTP verb to use to make the request.
        :type method:
            `unicode`
        :param url:
            The request URL.
        :type url:
            `unicode`
        :param headers:
            Headers to include with the request.
        :type headers:
            `dict`
        :param auto_session_renewal:
            Whether or not to automatically renew the session if the request fails due to an expired access token.
        :type auto_session_renewal:
            `bool`
        :param expect_json_response:
            Whether or not the response content should be json.
        :type expect_json_response:
            `bool`
        :param attempt_number:
            How many attempts at this request have already been tried. Used for exponential backoff calculations.
        :type attempt_number:
            `int`
        """
        files = kwargs.get('files')
        file_stream_positions = None
        if files:
            file_stream_positions = dict((name, file_tuple[1].tell()) for name, file_tuple in files.items())
        return self._make_request(
            method,
            url,
            headers,
            auto_session_renewal,
            expect_json_response,
            attempt_number,
            file_stream_positions=file_stream_positions,
            **kwargs
        )

    def _make_request(
            self,
            method,
            url,
            headers=None,
            auto_session_renewal=True,
            expect_json_response=True,
            attempt_number=0,
            **kwargs
    ):
        """
        Make a request to the Box API.

        :param method:
            The HTTP verb to use to make the request.
        :type method:
            `unicode`
        :param url:
            The request URL.
        :type url:
            `unicode`
        :param headers:
            Headers to include with the request.
        :type headers:
            `dict`
        :param auto_session_renewal:
            Whether or not to automatically renew the session if the request fails due to an expired access token.
        :type auto_session_renewal:
            `bool`
        :param expect_json_response:
            Whether or not the response content should be json.
        :type expect_json_response:
            `bool`
        :param attempt_number:
            How many attempts at this request have already been tried. Used for exponential backoff calculations.
        :type attempt_number:
            `int`
        """
        # Since there can be session renewal happening in the middle of preparing the request, it's important to be
        # consistent with the access_token being used in the request.
        access_token_will_be_used = self._oauth.access_token
        authorization_header = {'Authorization': 'Bearer {0}'.format(access_token_will_be_used)}
        if headers is None:
            headers = {}
        headers.update(authorization_header)

        # Reset stream positions to what they were when the request was made so the same data is sent even if this
        # is a retried attempt.
        files, file_stream_positions = kwargs.get('files'), kwargs.pop('file_stream_positions')
        if files and file_stream_positions:
            for name, position in file_stream_positions.items():
                files[name][1].seek(position)

        # send the request
        network_response = self._network_layer.request(
            method,
            url,
            access_token=access_token_will_be_used,
            headers=headers,
            **kwargs
        )

        network_response = self._retry_request_if_necessary(
            network_response,
            attempt_number,
            method,
            url,
            headers=headers,
            auto_session_renewal=auto_session_renewal,
            expect_json_response=expect_json_response,
            file_stream_positions=file_stream_positions,
            **kwargs
        )

        self._raise_on_unsuccessful_request(network_response, expect_json_response, method, url)

        return network_response

    def get(self, url, **kwargs):
        """Make a GET request to the Box API.

        :param url:
            The URL for the request.
        :type url:
            `unicode`
        """
        response = self._prepare_and_send_request('GET', url, **kwargs)
        return BoxResponse(response)

    def post(self, url, **kwargs):
        """Make a POST request to the Box API.

        :param url:
            The URL for the request.
        :type url:
            `unicode`
        """
        response = self._prepare_and_send_request('POST', url, **kwargs)
        return BoxResponse(response)

    def put(self, url, **kwargs):
        """Make a PUT request to the Box API.

        :param url:
            The URL for the request.
        :type url:
            `unicode`
        """
        response = self._prepare_and_send_request('PUT', url, **kwargs)
        return BoxResponse(response)

    def delete(self, url, **kwargs):
        """Make a DELETE request to the Box API.

        :param url:
            The URL for the request.
        :type url:
            `unicode`
        """
        if 'expect_json_response' not in kwargs:
            kwargs['expect_json_response'] = False
        response = self._prepare_and_send_request('DELETE', url, **kwargs)
        return BoxResponse(response)

    def options(self, url, **kwargs):
        """Make an OPTIONS request to the Box API.

        :param url:
            The URL for the request.
        :type url:
            `unicode`
        """
        response = self._prepare_and_send_request('OPTIONS', url, **kwargs)
        return BoxResponse(response)

    def request(self, method, url, **kwargs):
        """Make a request to the Box API.

        :param method:
            The HTTP verb for the request.
        :type method:
            `unicode`
        :param url:
            The URL for the request.
        :type url:
            `unicode`
        """
        response = self._prepare_and_send_request(method, url, **kwargs)
        return BoxResponse(response)



# coding: utf-8

from __future__ import unicode_literals

import six


__all__ = [
    'collaboration',
    'events',
    'file',
    'folder',
    'group',
    'group_membership',
    'search',
    'user',
]

if six.PY2:
    __all__ = [unicode.encode(x, 'utf-8') for x in __all__]



# coding: utf-8

from __future__ import unicode_literals
import json
import pytest
from boxsdk.object.metadata import MetadataUpdate


@pytest.fixture
def metadata_response():
    return {
        '$id': 'c79896a0-a33f-11e3-a5e2-0800200c9a66',
        '$type': 'properties',
        '$parent': 'file_552345101',
        'client_number': '820183',
        'client_name': 'Biomedical Corp',
        'case_reference': 'A83JAA',
        'case_type': 'Employment Litigation',
        'assigned_attorney': 'Francis Burke',
        'case_status': 'in-progress',
    }


@pytest.fixture(params=['enterprise', 'global'])
def metadata_scope(request):
    return request.param


@pytest.fixture
def metadata_template():
    return 'properties'


@pytest.mark.parametrize('success', [True, False])
def test_delete(mock_box_session, make_mock_box_request, test_file, metadata_scope, metadata_template, success):
    # pylint:disable=redefined-outer-name
    mock_box_session.delete.return_value, _ = make_mock_box_request(response_ok=success)
    metadata = test_file.metadata(metadata_scope, metadata_template)
    assert metadata.delete() is success
    mock_box_session.delete.assert_called_once_with(metadata.get_url())


def test_create(
        mock_box_session,
        make_mock_box_request,
        test_file,
        metadata_scope,
        metadata_template,
        metadata_response,
):
    # pylint:disable=redefined-outer-name
    mock_box_session.post.return_value, _ = make_mock_box_request(response=metadata_response)
    metadata = test_file.metadata(metadata_scope, metadata_template)
    response = metadata.create(metadata_response)
    assert response is metadata_response
    mock_box_session.post.assert_called_once_with(
        metadata.get_url(),
        data=json.dumps(metadata_response),
        headers={b'Content-Type': b'application/json'},
    )


def test_get(
        mock_box_session,
        make_mock_box_request,
        test_file,
        metadata_scope,
        metadata_template,
        metadata_response,
):
    # pylint:disable=redefined-outer-name
    mock_box_session.get.return_value, _ = make_mock_box_request(response=metadata_response)
    metadata = test_file.metadata(metadata_scope, metadata_template)
    response = metadata.get()
    assert response is metadata_response
    mock_box_session.get.assert_called_once_with(metadata.get_url())


@pytest.fixture
def metadata_update():
    update = MetadataUpdate()
    update.add('path', 'value')
    update.remove('path', 'value')
    update.test('path', 'value')
    update.update('path', 'value', 'value')
    return update


def test_update(
        mock_box_session,
        make_mock_box_request,
        test_file,
        metadata_scope,
        metadata_template,
        metadata_response,
        metadata_update,
):
    # pylint:disable=redefined-outer-name
    mock_box_session.put.return_value, _ = make_mock_box_request(response=metadata_response)
    metadata = test_file.metadata(metadata_scope, metadata_template)
    response = metadata.update(metadata_update)
    assert response is metadata_response
    mock_box_session.put.assert_called_once_with(
        metadata.get_url(),
        data=json.dumps(metadata_update.ops),
        headers={b'Content-Type': b'application/json-patch+json'},
    )



# coding: utf-8

from __future__ import unicode_literals
from datetime import datetime
import sqlalchemy
from sqlalchemy.orm import relationship, backref
import uuid
from test.functional.mock_box.db_model import DbModel


class FileModel(DbModel):
    """DB Model for Box files."""
    __tablename__ = 'box_file'

    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True, autoincrement=True)  # pylint:disable=invalid-name
    file_id = sqlalchemy.Column(sqlalchemy.String(32), default=lambda: uuid.uuid4().hex)
    name = sqlalchemy.Column(sqlalchemy.String(255))
    parent_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey('box_folder.id'))
    content = sqlalchemy.Column(sqlalchemy.LargeBinary)
    created_at = sqlalchemy.Column(sqlalchemy.DateTime, default=datetime.now)
    modified_at = sqlalchemy.Column(sqlalchemy.DateTime, onupdate=datetime.now)
    created_by_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey('box_user.id'))
    created_by = relationship('UserModel', foreign_keys=[created_by_id])
    owned_by_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey('box_user.id'))
    shared_link = relationship(
        'ShareFileModel',
        backref=backref('shared_file', remote_side=[id]),
        cascade='save-update, delete',
    )
    locks = relationship('LockModel', backref='item', cascade='save-update, delete')
    etag = sqlalchemy.Column(sqlalchemy.String(32), default=lambda: uuid.uuid4().hex)
    sequence_id = sqlalchemy.Column(sqlalchemy.Integer, nullable=False)
    sha1 = sqlalchemy.Column(sqlalchemy.String(40))
    size = sqlalchemy.Column(sqlalchemy.Integer)

    __mapper_args__ = {
        "version_id_col": sequence_id,
    }



# coding: utf-8

from __future__ import unicode_literals

from functools import partial
from io import IOBase
from mock import MagicMock, Mock, call
from numbers import Number
import pytest

from boxsdk.auth.oauth2 import OAuth2
from boxsdk.exception import BoxAPIException
from boxsdk.network.default_network import DefaultNetwork, DefaultNetworkResponse
from boxsdk.session.box_session import BoxSession, BoxResponse


@pytest.fixture
def box_session():
    mock_oauth = Mock(OAuth2)
    mock_oauth.access_token = 'fake_access_token'

    mock_network_layer = Mock(DefaultNetwork)

    return BoxSession(mock_oauth, mock_network_layer)


@pytest.mark.parametrize('test_method', [
    BoxSession.get,
    BoxSession.post,
    BoxSession.put,
    BoxSession.delete,
    BoxSession.options,
])
def test_box_session_handles_unauthorized_response(
        test_method,
        box_session,
        unauthorized_response,
        generic_successful_response,
        test_url,
):
    # pylint:disable=redefined-outer-name, protected-access
    mock_network_layer = box_session._network_layer
    mock_network_layer.request.side_effect = [unauthorized_response, generic_successful_response]

    box_response = test_method(box_session, url=test_url)
    assert box_response.status_code == 200


@pytest.mark.parametrize('test_method', [
    BoxSession.get,
    BoxSession.post,
    BoxSession.put,
    BoxSession.delete,
    BoxSession.options,
    partial(BoxSession.request, method='head'),
])
def test_box_session_retries_response_after_retry_after(
        test_method,
        box_session,
        retry_after_response,
        generic_successful_response,
        test_url,
):
    # pylint:disable=redefined-outer-name, protected-access
    mock_network_layer = box_session._network_layer
    mock_network_layer.request.side_effect = [retry_after_response, generic_successful_response]
    mock_network_layer.retry_after.side_effect = lambda delay, request, *args, **kwargs: request(*args, **kwargs)

    box_response = test_method(box_session, url=test_url)
    assert box_response.status_code == 200
    assert len(mock_network_layer.retry_after.call_args_list) == 1
    assert isinstance(mock_network_layer.retry_after.call_args[0][0], Number)
    assert mock_network_layer.retry_after.call_args[0][0] == float(retry_after_response.headers['Retry-After'])


@pytest.mark.parametrize('test_method', [
    BoxSession.get,
    BoxSession.post,
    BoxSession.put,
    BoxSession.delete,
    BoxSession.options,
    partial(BoxSession.request, method='head'),
])
def test_box_session_retries_request_after_server_error(
        test_method,
        box_session,
        server_error_response,
        generic_successful_response,
        test_url,
):
    # pylint:disable=redefined-outer-name, protected-access
    mock_network_layer = box_session._network_layer
    mock_network_layer.request.side_effect = [server_error_response, server_error_response, generic_successful_response]
    mock_network_layer.retry_after.side_effect = lambda delay, request, *args, **kwargs: request(*args, **kwargs)

    box_response = test_method(box_session, url=test_url)
    assert box_response.status_code == 200
    assert box_response.json() == generic_successful_response.json()
    assert box_response.ok == generic_successful_response.ok
    assert box_response.content == generic_successful_response.content
    assert len(mock_network_layer.retry_after.call_args_list) == 2
    assert isinstance(mock_network_layer.retry_after.call_args_list[0][0][0], Number)
    assert isinstance(mock_network_layer.retry_after.call_args_list[1][0][0], Number)
    assert mock_network_layer.retry_after.call_args_list[0][0][0] == 1
    assert mock_network_layer.retry_after.call_args_list[1][0][0] == 2


def test_box_session_seeks_file_after_retry(box_session, server_error_response, generic_successful_response, test_url):
    # pylint:disable=redefined-outer-name, protected-access
    mock_network_layer = box_session._network_layer
    mock_network_layer.request.side_effect = [server_error_response, generic_successful_response]
    mock_network_layer.retry_after.side_effect = lambda delay, request, *args, **kwargs: request(*args, **kwargs)
    mock_file_1, mock_file_2 = MagicMock(IOBase), MagicMock(IOBase)
    mock_file_1.tell.return_value = 0
    mock_file_2.tell.return_value = 3
    files = {'file': ('unused', mock_file_1), 'f2': ('unused', mock_file_2)}

    box_response = box_session.post(url=test_url, files=files)
    assert box_response.status_code == 200
    assert box_response.json() == generic_successful_response.json()
    assert box_response.ok == generic_successful_response.ok
    mock_file_1.tell.assert_called_once_with()
    mock_file_2.tell.assert_called_once_with()
    mock_file_1.seek.assert_called_with(0)
    assert mock_file_1.seek.call_count == 2
    assert mock_file_1.seek.has_calls(call(0) * 2)
    mock_file_2.seek.assert_called_with(3)
    assert mock_file_2.seek.call_count == 2
    assert mock_file_2.seek.has_calls(call(3) * 2)


def test_box_session_raises_for_non_json_response(box_session, non_json_response, test_url):
    # pylint:disable=redefined-outer-name, protected-access
    mock_network_layer = box_session._network_layer
    mock_network_layer.request.side_effect = [non_json_response]

    with pytest.raises(BoxAPIException):
        box_session.get(url=test_url)


def test_box_session_raises_for_failed_response(box_session, bad_network_response, test_url):
    # pylint:disable=redefined-outer-name, protected-access
    mock_network_layer = box_session._network_layer
    mock_network_layer.request.side_effect = [bad_network_response]

    with pytest.raises(BoxAPIException):
        box_session.get(url=test_url)


def test_box_session_raises_for_failed_non_json_response(box_session, failed_non_json_response, test_url):
    # pylint:disable=redefined-outer-name, protected-access
    mock_network_layer = box_session._network_layer
    mock_network_layer.request.side_effect = [failed_non_json_response]

    with pytest.raises(BoxAPIException):
        box_session.get(url=test_url, expect_json_response=False)


def test_box_response_properties_pass_through_to_network_response_properties():
    mock_network_response = Mock(DefaultNetworkResponse)
    box_result = BoxResponse(mock_network_response)
    assert box_result.json() == mock_network_response.json()
    assert box_result.content == mock_network_response.content
    assert box_result.ok == mock_network_response.ok
    assert box_result.status_code == mock_network_response.status_code
    assert box_result.network_response == mock_network_response



# coding: utf-8

from __future__ import unicode_literals
import json
from boxsdk.object.base_endpoint import BaseEndpoint


class MetadataUpdate(object):
    """
    Helper class for updating Box metadata.
    See https://developers.box.com/metadata-api/#update for more details.
    See http://jsonpatch.com/ for details about JSON patch.
    """
    def __init__(self):
        self._ops = []

    @property
    def ops(self):
        """
        Get a list of json patch operations in this update.

        :return:
            The list of json patch operations in this update.
        :rtype:
            `list` of `dict`
        """
        return self._ops

    def add(self, path, value):
        """
        Insert an add operation to this metadata update.

        :param path:
            JSON pointer specifying where to add the new value.
        :type path:
            `unicode`
        :param value:
            The value to add to the metadata document.
        :type value:
            `unicode`
        """
        self._ops.append({'op': 'add', 'path': path, 'value': value})

    def remove(self, path, old_value=None):
        """
        Insert a remove operation to this metadata update.

        :param path:
            JSON pointer specifying where to remove the value from.
        :type path:
            `unicode`
        :param old_value:
            If specified, only remove the key/value pair if the current value is equal to oldValue.
        :type old_value:
            `unicode`
        """
        if old_value is not None:
            self._ops.append({'op': 'test', 'path': path, 'value': old_value})
        self._ops.append({'op': 'remove', 'path': path})

    def update(self, path, value, old_value=None):
        """
        Insert an update operation to this metadata update.

        :param path:
            JSON pointer specifying where the value is in the metadata document that should be updated.
        :type path:
            `unicode`
        :param value:
            The updated value.
        :type value:
            `unicode`
        :param old_value:
            If specified, only update the key/value pair if the current value is equal to oldValue.
        :type old_value:
            `unicode`
        """
        if old_value is not None:
            self._ops.append({'op': 'test', 'path': path, 'value': old_value})
        self._ops.append({'op': 'replace', 'path': path, 'value': value})

    def test(self, path, value):
        """
        Insert a test operation to this metadata update.
        A test operation can invalidate the following operation if the value at the specified path does not match
        the supplied value.

        :param path:
            JSON pointer specifying where the value is in the metadata document to test.
        :type path:
            `unicode`
        :param value:
            The value to match against.
        :type value:
            `unicode`
        """
        self._ops.append({'op': 'test', 'path': path, 'value': value})


class Metadata(BaseEndpoint):
    def __init__(self, session, box_object, scope, template):
        """
        :param session:
            The Box session used to make requests.
        :type session:
            :class:`BoxSession`
        :param box_object:
            The Box object this metadata instance will be associated with.
        :type box_object:
            :class:`BaseObject`
        :param scope:
            Scope of the metadata. Must be either 'global' or 'enterprise'.
        :type scope:
            `unicode`
        :param template:
            The name of the metadata template. See https://developers.box.com/metadata-api/#basics for more details.
        :type template:
            `unicode`
        """
        super(Metadata, self).__init__(session)
        self._object = box_object
        self._scope = scope
        self._template = template

    def get_url(self, *args):
        return self._object.get_url('metadata', self._scope, self._template)

    @staticmethod
    def start_update():
        """
        Get a :class:`MetadataUpdate` for use with the :meth:`update` method.

        :return:
            A metadata update object that can be used to update this metadata object.
        :rtype:
            :class:`MetadataUpdate`
        """
        return MetadataUpdate()

    def update(self, metadata_update):
        """
        Update the key/value pairs associated with this metadata object.
        See https://developers.box.com/metadata-api/#update for more details.

        :param metadata_update:
            A metadata update object containing the changes that should be made to the metadata.
        :type metadata_update:
            :class:`MetadataUpdate`
        :return:
            A dictionary containing the updated key/value pairs for this metadata object.
        :rtype:
            :class:`Metadata`
        """
        return self._session.put(
            self.get_url(),
            data=json.dumps(metadata_update.ops),
            headers={b'Content-Type': b'application/json-patch+json'},
        ).json()

    def get(self):
        """
        Get the key/value pairs that make up this metadata instance.

        :return:
            A dictionary containing the key/value pairs for this metadata object.
        :rtype:
            :class:`Metadata`
        """
        return self._session.get(self.get_url()).json()

    def delete(self):
        """
        Delete the metadata object.

        :returns:
            Whether or not the delete was successful.
        :rtype:
            `bool`
        """
        return self._session.delete(self.get_url()).ok

    def create(self, metadata):
        """
        Create the metadata instance on Box. If the instance already exists, use :meth:`update` instead.

        :param metadata:
            The key/value pairs to be stored in this metadata instance on Box.
        :type metadata:
            `dict`
        :return:
            A dictionary containing the key/value pairs for this metadata object.
        :rtype:
            :class:`Metadata`
        """
        return self._session.post(
            self.get_url(),
            data=json.dumps(metadata),
            headers={b'Content-Type': b'application/json'},
        ).json()



# coding: utf-8

from __future__ import unicode_literals

from os.path import dirname, join
import sys
from sys import version_info

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand


CLASSIFIERS = [
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: Apache Software License',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2.6',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: Implementation :: CPython',
    'Programming Language :: Python :: Implementation :: PyPy',
    'Operating System :: OS Independent',
    'Operating System :: POSIX',
    'Operating System :: Microsoft :: Windows',
    'Operating System :: MacOS :: MacOS X',
    'Topic :: Software Development :: Libraries :: Python Modules',
]


class PyTest(TestCommand):
    user_options = [(b'pytest-args=', b'a', b"Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = None

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # Do the import here, once the eggs are loaded.
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


def main():
    base_dir = dirname(__file__)
    install_requires = ['requests>=2.4.3', 'six>=1.4.0']
    if version_info < (3, 4):
        install_requires.append('enum34>=1.0.4')
    elif version_info < (2, 7):
        install_requires.append('ordereddict>=1.1')
    setup(
        name='boxsdk',
        version='1.1.7',
        description='Official Box Python SDK',
        long_description=open(join(base_dir, 'README.rst')).read(),
        author='Box',
        author_email='oss@box.com',
        url='http://opensource.box.com',
        packages=find_packages(exclude=['demo', 'docs', 'test']),
        install_requires=install_requires,
        tests_require=['pytest', 'pytest-xdist', 'mock', 'sqlalchemy', 'bottle', 'jsonpatch'],
        cmdclass={'test': PyTest},
        classifiers=CLASSIFIERS,
        keywords='box oauth2 sdk',
        license=open(join(base_dir, 'LICENSE')).read(),
    )


if __name__ == '__main__':
    main()



# coding: utf-8

from __future__ import unicode_literals
from abc import ABCMeta
import json

import six

from boxsdk.object.base_endpoint import BaseEndpoint
from boxsdk.util.translator import Translator


class ObjectMeta(ABCMeta):
    """
    Metaclass for Box API objects. Registers classes so that API responses can be translated to the correct type.
    Relies on the _item_type field defined on the classes to match the type property of the response json.
    But the type-class mapping will only be registered if the module of the class is imported.
    So it's also important to add the module name to __all__ in object/__init__.py.
    """
    def __init__(cls, name, bases, attrs):
        super(ObjectMeta, cls).__init__(name, bases, attrs)
        item_type = attrs.get('_item_type', None)
        if item_type is not None:
            Translator().register(item_type, cls)


@six.add_metaclass(ObjectMeta)
class BaseObject(BaseEndpoint):
    """
    A Box API endpoint for interacting with a Box object.
    """
    _item_type = None

    def __init__(self, session, object_id, response_object=None):
        """
        :param session:
            The Box session used to make requests.
        :type session:
            :class:`BoxSession`
        :param object_id:
            The Box ID for the object.
        :type object_id:
            `unicode`
        :param response_object:
            The Box API response representing the object.
        :type response_object:
            :class:`BoxResponse`
        """
        super(BaseObject, self).__init__(session)
        self._object_id = object_id
        self._response_object = response_object or {}

    def __getattr__(self, item):
        """Base class override. Try to get the attribute from the API response object."""
        return self._response_object[item]

    def __getitem__(self, item):
        """Base class override. Try to get the attribute from the API response object."""
        return self._response_object[item]

    def get_url(self, *args):
        """
        Base class override.
        Return the given object's URL, appending any optional parts as specified by args.
        """
        return super(BaseObject, self).get_url('{0}s'.format(self._item_type), self._object_id, *args)

    def get_type_url(self):
        """
        Return the URL for type of the given resource.

        :rtype:
            `unicode`
        """
        return super(BaseObject, self).get_url('{0}s'.format(self._item_type))

    @property
    def object_id(self):
        """Return the Box ID for the object.

        :rtype:
            `unicode`
        """
        return self._object_id

    def get(self, fields=None, headers=None):
        """
        Get information about the object, specified by fields. If fields is None, return the default fields.

        :param fields:
            List of fields to request.
        :type fields:
            `Iterable` of `unicode`
        :param headers:
            Additional headers to send with the request.
        :type headers:
            `dict`
        :return:
            An object of the same type that has the requested information.
        :rtype:
            :class:`BaseObject`
        """
        url = self.get_url()
        params = {'fields': ','.join(fields)} if fields else None
        box_response = self._session.get(url, params=params, headers=headers)
        return self.__class__(self._session, self._object_id, box_response.json())

    def update_info(self, data, params=None, headers=None, **kwargs):
        """Update information about this object.

        Send a PUT to the object's base endpoint to modify the provided
        attributes.

        :param data:
            The updated information about this object.
            Must be JSON serializable.
            Update the object attributes in data.keys(). The semantics of the
            values depends on the the type and attributes of the object being
            updated. For details on particular semantics, refer to the Box
            developer API documentation <developers.box.com/docs>.
        :type data:
            `dict`
        :param params:
            (optional) Query string parameters for the request.
        :type params:
            `dict` or None
        :param headers:
            (optional) Extra HTTP headers for the request.
        :type headers:
            `dict` or None
        :param kwargs:
            Optional arguments that ``put`` takes.
        :return:
            The updated object.
            Return a new object of the same type, without modifying the
            original object passed as self.
            Construct the new object with all the default attributes that are
            returned from the endpoint.
        :rtype:
            :class:`BaseObject`
        """
        url = self.get_url()
        box_response = self._session.put(url, data=json.dumps(data), params=params, headers=headers, **kwargs)
        response = box_response.json()
        return self.__class__(
            session=self._session,
            object_id=self._object_id,
            response_object=response,
        )

    def delete(self, params=None, headers=None):
        """ Delete the object.

        :param params:
            Additional parameters to send with the request. Can be None
        :type params:
            `dict` or None
        :param headers:
            Any customer headers to send with the request. Can be None
        :type headers:
            `dict` or None
        :returns:
            Whether or not the delete was successful.
        :rtype:
            `bool`
        :raises:
            :class:`BoxAPIException` in case of unexpected errors.
        """
        url = self.get_url()
        # ??? There's a question about why params forces a default to {}, while headers doesn't. Looking for
        # confirmation that the below is correct.
        box_response = self._session.delete(url, expect_json_response=False, params=params or {}, headers=headers)
        return box_response.ok

    def __eq__(self, other):
        """Base class override. Equality is determined by object id."""
        return self._object_id == other.object_id

    def _paging_wrapper(self, url, starting_index, limit, factory=None):
        """
        Helper function that turns any paging API into a generator that transparently implements the paging for
        the caller.

        A caller wanting to implement their own paging can do so by managing the starting_index & limit params,
        and never iterating over more than 'limit' items per call. For example:

            first_ten = list(itertools.islice(_paging_wrapper(..., 0, 10, ...), 10))
            second_ten = list(itertools.islice(_paging_wrapper(..., 10, 10, ...), 10))
            third_ten = list(itertools.islice(_paging_wrapper(..., 20, 10, ...), 10))
            ...
        When one of the lists has less than 10 items... the end has been reached.

        Caveat: any hidden items (see the Box Developer API for more details) will render the above
        inaccurate. Hidden results will lead the above get_slice() code to trigger API calls at non-expected places.

        :param starting_index:
            The index at which to begin.
        :type starting_index:
            `int`
        :param limit:
            The maximum number of items to return in a page.
        :type limit:
            `int`
        :param factory:
            A callable factory method which creates the object instances. Signature should match the __init__
            signature of BaseObject. If no factory is given then the Translator factory is used.
        :type factory:
            `callable` or None
        :returns:
            A generator of 3-tuples. Each tuple contains:
            1) An instance returned by the given factory callable.
            2) The number of objects in the current page.
            3) Index the current instance in the current page.
        :rtype:
            `generator` of `tuple` of (varies, `int`, `int`)
        """
        current_index = starting_index

        while True:
            params = {'limit': limit, 'offset': current_index}
            box_response = self._session.get(url, params=params)
            response = box_response.json()

            current_page_size = len(response['entries'])
            for index_in_current_page, item in enumerate(response['entries']):
                instance_factory = factory
                if not instance_factory:
                    instance_factory = Translator().translate(item['type'])
                instance = instance_factory(self._session, item['id'], item)
                yield instance, current_page_size, index_in_current_page

            current_index += limit
            if current_index >= response['total_count']:
                break



# coding: utf-8

from __future__ import unicode_literals
import pytest
import json
from boxsdk.object.file import File
from boxsdk.object.search import MetadataSearchFilters, MetadataSearchFilter


@pytest.fixture
def search_query():
    return 'myquery'


@pytest.fixture(params=(1, 20, 100))
def search_limit(request):
    return request.param


@pytest.fixture(params=(0, 10))
def search_offset(request):
    return request.param


@pytest.fixture(params=(None, 'file', 'folder'))
def search_result_type(request):
    return request.param


@pytest.fixture(params=(None, ('name',), ('name', 'description')))
def search_content_types(request):
    return request.param


@pytest.fixture
def search_value_based_filters():
    metadata_filters = MetadataSearchFilters()
    metadata_filter = MetadataSearchFilter(template_key='mytemplate', scope='enterprise')
    metadata_filter.add_value_based_filter(field_key='myfield', value='myvalue')
    metadata_filters.add_filter(metadata_filter)
    return metadata_filters


@pytest.fixture(params=(
    {'gt_value': 'mygtvalue'},
    {'lt_value': 'myltvalue'},
    {'gt_value': 'mygtvalue', 'lt_value': 'myltvalue'}
))
def search_range_filters(request):
    metadata_filters = MetadataSearchFilters()
    metadata_filter = MetadataSearchFilter(template_key='mytemplate', scope='enterprise')
    filter_params = {'field_key': 'myfield'}
    filter_params.update(request.param)
    metadata_filter.add_range_filter(**filter_params)
    metadata_filters.add_filter(metadata_filter)
    return metadata_filters


@pytest.fixture
def search_entries():
    return [
        {'id': '1234', 'type': 'file'}
    ]


@pytest.fixture
def search_response():
    return {
        'entries': search_entries(),
        'total_count': 0,
        'limit': 20,
        'offset': 0
    }


class Matcher(object):
    def __init__(self, compare, some_obj):
        self.compare = compare
        self.some_obj = some_obj

    def __eq__(self, other):
        return self.compare(self.some_obj, other)


def compare_params(self, other):
    if not isinstance(self, dict) or not isinstance(other, dict):
        return False
    for key in self:
        # We need to ensure that the JSON-encoded mdfilters matches regardless of key order
        if key == 'mdfilters':
            if json.loads(self['mdfilters']) != json.loads(other['mdfilters']):
                return False
        # For other keys, just ensure that they are equal
        elif key in ('type', 'content_types'):
            return self[key] is None or self[key] == other[key]
        else:
            if self[key] != other[key]:
                return False
    return True


def test_search_with_value_based_filters(
        mock_box_session,
        make_mock_box_request,
        test_search,
        search_query,
        search_limit,
        search_offset,
        search_value_based_filters,
        search_response,
        search_entries,
        search_result_type,
        search_content_types,
):
    # pylint:disable=redefined-outer-name
    mock_box_session.get.return_value, _ = make_mock_box_request(response=search_response)
    response = test_search.search(
        search_query,
        limit=search_limit,
        offset=search_offset,
        metadata_filters=search_value_based_filters,
        result_type=search_result_type,
        content_types=search_content_types,
    )
    assert response == [File(mock_box_session, search_entry['id'], search_entry) for search_entry in search_entries]

    mock_box_session.get.assert_called_once_with(
        test_search.get_url(),
        params=Matcher(compare_params, {
            'query': search_query,
            'limit': search_limit,
            'mdfilters': json.dumps(search_value_based_filters.as_list()),
            'offset': search_offset,
            'type': search_result_type,
            'content_types': ','.join(search_content_types) if search_content_types else search_content_types,
        })
    )


def test_search_with_range_filters(
        mock_box_session,
        make_mock_box_request,
        test_search,
        search_query,
        search_limit,
        search_offset,
        search_range_filters,
        search_response,
        search_entries,
        search_result_type,
        search_content_types,
):
    # pylint:disable=redefined-outer-name
    mock_box_session.get.return_value, _ = make_mock_box_request(response=search_response)
    response = test_search.search(
        search_query,
        limit=search_limit,
        offset=search_offset,
        metadata_filters=search_range_filters,
        result_type=search_result_type,
        content_types=search_content_types,
    )
    assert response == [File(mock_box_session, search_entry['id'], search_entry) for search_entry in search_entries]

    mock_box_session.get.assert_called_once_with(
        test_search.get_url(),
        params=Matcher(compare_params, {
            'query': search_query,
            'limit': search_limit,
            'mdfilters': json.dumps(search_range_filters.as_list()),
            'offset': search_offset,
            'type': search_result_type,
            'content_types': ','.join(search_content_types) if search_content_types else search_content_types,
        })
    )


def test_range_filter_without_gt_and_lt_will_fail_validation():
    metadata_filter = MetadataSearchFilter(template_key='mytemplate', scope='enterprise')
    with pytest.raises(ValueError):
        metadata_filter.add_range_filter(field_key='mykey')



# coding: utf-8

from __future__ import unicode_literals
import json
import pytest


@pytest.fixture(params=('file', 'folder'))
def test_item_and_response(test_file, test_folder, mock_file_response, mock_folder_response, request):
    if request.param == 'file':
        return test_file, mock_file_response
    elif request.param == 'folder':
        return test_folder, mock_folder_response


def test_update_info(test_item_and_response, mock_box_session, etag, if_match_header):
    # pylint:disable=redefined-outer-name, protected-access
    test_item, mock_item_response = test_item_and_response
    expected_url = test_item.get_url()
    mock_box_session.put.return_value = mock_item_response
    data = {'foo': 'bar', 'baz': {'foo': 'bar'}, 'num': 4}
    update_response = test_item.update_info(data, etag=etag)
    mock_box_session.put.assert_called_once_with(expected_url, data=json.dumps(data), headers=if_match_header, params=None)
    assert isinstance(update_response, test_item.__class__)
    assert update_response.object_id == test_item.object_id


def test_rename_item(test_item_and_response, mock_box_session):
    # pylint:disable=redefined-outer-name, protected-access
    test_item, mock_item_response = test_item_and_response
    expected_url = test_item.get_url()
    mock_box_session.put.return_value = mock_item_response
    rename_response = test_item.rename('new name')
    mock_box_session.put.assert_called_once_with(expected_url, data=json.dumps({'name': 'new name'}), params=None, headers=None)
    assert isinstance(rename_response, test_item.__class__)


def test_copy_item(test_item_and_response, mock_box_session, test_folder, mock_object_id):
    # pylint:disable=redefined-outer-name, protected-access
    test_item, mock_item_response = test_item_and_response
    expected_url = test_item.get_url('copy')
    mock_box_session.post.return_value = mock_item_response
    copy_response = test_item.copy(test_folder)
    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps({'parent': {'id': mock_object_id}}))
    assert isinstance(copy_response, test_item.__class__)


def test_move_item(test_item_and_response, mock_box_session, test_folder, mock_object_id):
    # pylint:disable=redefined-outer-name, protected-access
    test_item, mock_item_response = test_item_and_response
    expected_url = test_item.get_url()
    mock_box_session.put.return_value = mock_item_response
    move_response = test_item.move(test_folder)
    mock_box_session.put.assert_called_once_with(expected_url, data=json.dumps({'parent': {'id': mock_object_id}}), params=None, headers=None)
    assert isinstance(move_response, test_item.__class__)


@pytest.mark.parametrize('access,expected_access_data', [(None, {}), ('open', {'access': 'open'})])
def test_get_shared_link(
        test_item_and_response,
        mock_box_session,
        access,
        expected_access_data,
        test_url,
        etag,
        if_match_header,
):
    # pylint:disable=redefined-outer-name, protected-access
    test_item, _ = test_item_and_response
    expected_url = test_item.get_url()
    mock_box_session.put.return_value.json.return_value = {'shared_link': {'url': test_url}}
    url = test_item.get_shared_link(access, etag=etag)
    mock_box_session.put.assert_called_once_with(
        expected_url,
        data=json.dumps({'shared_link': expected_access_data}),
        headers=if_match_header,
        params=None,
    )
    assert url == test_url


def test_remove_shared_link(test_item_and_response, mock_box_session, etag, if_match_header):
    # pylint:disable=redefined-outer-name, protected-access
    test_item, _ = test_item_and_response
    expected_url = test_item.get_url()
    mock_box_session.put.return_value.json.return_value = {'shared_link': None}
    removed = test_item.remove_shared_link(etag=etag)
    mock_box_session.put.assert_called_once_with(
        expected_url,
        data=json.dumps({'shared_link': None}),
        headers=if_match_header,
        params=None,
    )
    assert removed is True


@pytest.mark.parametrize('fields', (None, ['name', 'created_at']))
def test_get(test_item_and_response, mock_box_session, fields, mock_object_id, etag, if_none_match_header):
    # pylint:disable=redefined-outer-name, protected-access
    test_item, mock_item_response = test_item_and_response
    expected_url = test_item.get_url()
    mock_box_session.get.return_value = mock_item_response
    expected_params = {'fields': ','.join(fields)} if fields else None
    info = test_item.get(fields, etag=etag)
    mock_box_session.get.assert_called_once_with(expected_url, params=expected_params, headers=if_none_match_header)
    assert isinstance(info, test_item.__class__)
    assert info.id == mock_object_id



# coding: utf-8

from __future__ import unicode_literals


def test_too_many_requests_causes_retry(box_client, mock_box, monkeypatch):
    monkeypatch.setattr(mock_box, 'RATE_LIMIT_THRESHOLD', 1)
    box_client.folder(0).get()
    box_client.folder(0).get()
    assert len(mock_box.requests) == 6  # 3 auth requests, 2 real requests, and a retry



# coding: utf-8

from __future__ import unicode_literals
import sqlalchemy
import uuid
from test.functional.mock_box.db_model import DbModel


class ShareModel(DbModel):
    """DB Model for Box shared links."""
    __tablename__ = 'box_share'

    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True, autoincrement=True)  # pylint:disable=invalid-name
    share_id = sqlalchemy.Column(sqlalchemy.String(32), default=lambda: uuid.uuid4().hex)
    url = sqlalchemy.Column(sqlalchemy.String(50))


class ShareFileModel(ShareModel):
    """DB Model for Box shared links to files."""
    file_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey('box_file.id'))


class ShareFolderModel(ShareModel):
    """DB Model for Box shared links to folders."""
    folder_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey('box_folder.id'))



# coding: utf-8

from __future__ import unicode_literals
import json

from .config import API
from .session.box_session import BoxSession
from .network.default_network import DefaultNetwork
from .object.user import User
from .object.folder import Folder
from .object.search import Search
from .object.events import Events
from .object.file import File
from .object.group import Group
from .object.group_membership import GroupMembership
from .util.translator import Translator


class Client(object):

    def __init__(self, oauth, network_layer=None):
        """
        :param oauth:
            OAuth2 object used by the session to authorize requests.
        :type oauth:
            :class:`OAuth2`
        :param network_layer:
            The Network layer to use. If none is provided then an instance of :class:`DefaultNetwork` will be used.
        :type network_layer:
            :class:`Network`
        """
        network_layer = network_layer or DefaultNetwork()
        self._session = BoxSession(oauth=oauth, network_layer=network_layer)

    def folder(self, folder_id):
        """
        Initialize a :class:`Folder` object, whose box id is folder_id.

        :param folder_id:
            The box id of the :class:`Folder` object. Can use '0' to get the root folder on Box.
        :type folder_id:
            `unicode`
        :return:
            A :class:`Folder` object with the given folder id.
        :rtype:
            :class:`Folder`
        """
        return Folder(session=self._session, object_id=folder_id)

    def file(self, file_id):
        """
        Initialize a :class:`File` object, whose box id is file_id.

        :param file_id:
            The box id of the :class:`File` object.
        :type file_id:
            `unicode`
        :return:
            A :class:`File` object with the given file id.
        :rtype:
            :class:`File`
        """
        return File(session=self._session, object_id=file_id)

    def user(self, user_id='me'):
        """
        Initialize a :class:`User` object, whose box id is user_id.

        :param user_id:
            The user id of the :class:`User` object. Can use 'me' to get the User for the current/authenticated user.
        :type user_id:
            `unicode`
        :return:
            A :class:`User` object with the given id.
        :rtype:
            :class:`User`
        """
        return User(session=self._session, object_id=user_id)

    def group(self, group_id):
        """
        Initialize a :class:`Group` object, whose box id is group_id.

        :param group_id:
            The box id of the :class:`Group` object.
        :type group_id:
            `unicode`
        :return:
            A :class:`Group` object with the given group id.
        :rtype:
            :class:`Group`
        """
        return Group(session=self._session, object_id=group_id)

    def users(self):
        """
        Get a list of all users for the Enterprise along with their user_id, public_name, and login.

        :return:
            The list of all users in the enterprise.
        :rtype:
            `list` of :class:`User`
        """
        url = '{0}/users'.format(API.BASE_API_URL)
        box_response = self._session.get(url)
        response = box_response.json()
        return [User(self._session, item['id'], item) for item in response['entries']]

    def search(self, query, limit, offset, ancestor_folders=None, file_extensions=None, metadata_filters=None):
        """
        Search Box for items matching the given query.

        :param query:
            The string to search for.
        :type query:
            `unicode`
        :param limit:
            The maximum number of items to return.
        :type limit:
            `int`
        :param offset:
            The search result at which to start the response.
        :type offset:
            `int`
        :param ancestor_folders:
            Folder ids to limit the search to.
        :type ancestor_folders:
            `iterable` of :class:`Folder`
        :param file_extensions:
            File extensions to limit the search to.
        :type file_extensions:
            `iterable` of `unicode`
        :param metadata_filters:
            Filters used for metadata search
        :type metadata_filters:
            :class:`MetadataSearchFilters`
        :return:
            A list of items that match the search query.
        :rtype:
            `list` of :class:`Item`
        """
        return Search(self._session).search(query=query,
                                            limit=limit,
                                            offset=offset,
                                            ancestor_folders=ancestor_folders,
                                            file_extensions=file_extensions,
                                            metadata_filters=metadata_filters)

    def events(self):
        """
        Get an events object that can get the latest events from Box or set up a long polling event subscription.
        """
        return Events(self._session)

    def group_membership(self, group_membership_id):
        """
        Initialize a :class:`GroupMembership` object, whose box id is group_membership_id.

        :param group_membership_id:
            The box id of the :class:`GroupMembership` object.
        :type group_membership_id:
            `unicode`
        :return:
            A :class:`GroupMembership` object with the given membership id.
        :rtype:
            :class:`GroupMembership`
        """
        return GroupMembership(session=self._session, object_id=group_membership_id)

    def groups(self):
        """
        Get a list of all groups for the current user.

        :return:
            The list of all groups.
        :rtype:
            `list` of :class:`Group`
        """
        url = '{0}/groups'.format(API.BASE_API_URL)
        box_response = self._session.get(url)
        response = box_response.json()
        return [Group(self._session, item['id'], item) for item in response['entries']]

    def create_group(self, name):
        """
        Create a group with the given name.

        :param name:
            The name of the group.
        :type name:
            `unicode`
        :return:
            The newly created Group.
        :rtype:
            :class:`Group`
        :raises:
            :class:`BoxAPIException` if current user doesn't have permissions to create a group.
        """
        url = '{0}/groups'.format(API.BASE_API_URL)
        body_attributes = {
            'name': name,
        }
        box_response = self._session.post(url, data=json.dumps(body_attributes))
        response = box_response.json()
        return Group(self._session, response['id'], response)

    def get_shared_item(self, shared_link, password=None):
        """
        Get information about a Box shared link. https://developers.box.com/docs/#shared-items

        :param shared_link:
            The shared link.
        :type shared_link:
            `unicode`
        :param password:
            The password for the shared link.
        :type password:
            `unicode`
        :return:
            The item referred to by the shared link.
        :rtype:
            :class:`Item`
        :raises:
            :class:`BoxAPIException` if current user doesn't have permissions to view the shared link.
        """
        shared_link_password = '&shared_link_password={0}'.format(password) if password is not None else ''
        box_api_header = 'shared_link={0}{1}'.format(shared_link, shared_link_password)
        response = self.make_request(
            'GET',
            '{0}/shared_items'.format(API.BASE_API_URL),
            headers={'BoxApi': box_api_header},
        ).json()
        return Translator().translate(response['type'])(self._session, response['id'], response)

    def make_request(self, method, url, **kwargs):
        """
        Make an authenticated request to the Box API.

        :param method:
            The HTTP verb to use for the request.
        :type method:
            `unicode`
        :param url:
            The URL for the request.
        :type url:
            `unicode`
        :return:
            The network response for the given request.
        :rtype:
            :class:`BoxResponse`
        :raises:
            :class:`BoxAPIException`
        """
        return self._session.request(method, url, **kwargs)



# coding: utf-8

from __future__ import unicode_literals
from six import PY2


class BoxException(Exception):
    """
    Base class exception for all errors raised from the SDK.
    """
    def __str__(self):
        return self.__unicode__().encode('utf-8') if PY2 else self.__unicode__()


class BoxNetworkException(BoxException):
    """
    Exception raised from the network layer.
    """
    pass


class BoxAPIException(BoxException):
    """
    Exception raised from the box session layer.
    """
    def __init__(self, status, code=None, message=None, request_id=None, headers=None, url=None, method=None, context_info=None):
        """
        :param status:
            HTTP status code of the failed response
        :type status:
            `int`
        :param code:
            The 'code' field of the failed response
        :type code:
            `unicode`
        :param message:
            A message to associate with the exception, e.g. 'message' field of the json in the failed response
        :type message:
            `unicode`
        :param request_id:
            The 'request_id' field of the json in the failed response
        :type request_id:
            `unicode`
        :param headers:
            The HTTP headers in the failed response
        :type headers:
            `dict`
        :param url:
            The url which raised the exception
        :type url:
            `unicode`
        :param method:
            The HTTP verb used to make the request.
        :type method:
            `unicode`
        :param context_info:
            The context_info returned in the failed response.
        :type context_info:
            `dict`
        """
        super(BoxAPIException, self).__init__()
        self._status = status
        self._code = code
        self._message = message
        self._request_id = request_id
        self._headers = headers
        self._url = url
        self._method = method
        self._context_info = context_info

    def __unicode__(self):
        return '\nMessage: {0}\nStatus: {1}\nCode: {2}\nRequest id: {3}\nHeaders: {4}\nURL: {5}\nMethod: {6}\nContext info: {7}'.format(
            self._message,
            self._status,
            self._code,
            self._request_id,
            self._headers,
            self._url,
            self._method,
            self._context_info,
        )

    @property
    def status(self):
        """
        The status code of the network response that is responsible for the exception.
        :rtype: `int`
        """
        return self._status

    @property
    def code(self):
        """
        The explanation of the status code of the network response that is responsible for the exception.
        :rtype: `int`
        """
        return self._code

    @property
    def message(self):
        return self._message

    @property
    def request_id(self):
        """
        The id the network request that is responsible for the exception.
        :rtype: `unicode`
        """
        return self._request_id

    @property
    def url(self):
        """
        The URL of the network request that is responsible for the exception.
        :rtype: `unicode`
        """
        return self._url

    @property
    def method(self):
        """
        The HTTP verb of the request that is responsible for the exception.
        :rtype: `unicode`
        """
        return self._method

    @property
    def context_info(self):
        """
        The context_info returned in the failed response.
        :rtype: `dict`
        """
        return self._context_info


class BoxOAuthException(BoxException):
    """
    Exception raised during auth.
    """
    def __init__(self, status, message=None, url=None, method=None):
        """
        :param status:
            HTTP status code of the auth response
        :type status:
            `int`
        :param message:
            A message to associate with the exception, e.g. HTTP content of the auth response
        :type message:
            `unicode`
        :param url:
            The url which raised the exception
        :type url:
            `unicode`
        :param method:
            The HTTP verb used to make the request.
        :type method:
            `unicode`
        """
        super(BoxOAuthException, self).__init__()
        self._status = status
        self._message = message
        self._url = url
        self._method = method

    def __unicode__(self):
        return '\nMessage: {0}\nStatus: {1}\nURL: {2}\nMethod: {3}'.format(
            self._message,
            self._status,
            self._url,
            self._method,
        )



# coding: utf-8

from __future__ import unicode_literals
import json
from mock import Mock
import pytest
from requests.packages.urllib3.exceptions import TimeoutError
from boxsdk.network.default_network import DefaultNetworkResponse
from boxsdk.object.events import Events
from boxsdk.session.box_session import BoxResponse


@pytest.fixture()
def test_events(mock_box_session):
    return Events(mock_box_session)


@pytest.fixture()
def final_stream_position():
    return 1348790499820


@pytest.fixture()
def initial_stream_position():
    return 1348790499819


@pytest.fixture()
def empty_events_response(final_stream_position):
    # pylint:disable=redefined-outer-name
    mock_box_response = Mock(BoxResponse)
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_box_response.network_response = mock_network_response
    mock_box_response.json.return_value = mock_json = {'next_stream_position': final_stream_position, 'entries': []}
    mock_box_response.content = json.dumps(mock_json).encode()
    mock_box_response.status_code = 200
    mock_box_response.ok = True
    return mock_box_response


@pytest.fixture()
def long_poll_url(test_url):
    return test_url


@pytest.fixture()
def retry_timeout():
    return 610


@pytest.fixture()
def options_response(long_poll_url, retry_timeout, make_mock_box_request):
    # pylint:disable=redefined-outer-name
    mock_box_response, _ = make_mock_box_request(
        response={'entries': [{'url': long_poll_url, 'retry_timeout': retry_timeout}]},
    )
    return mock_box_response


@pytest.fixture()
def new_change_long_poll_response(make_mock_box_request):
    mock_box_response, _ = make_mock_box_request(
        response={'message': 'new_change'},
    )
    return mock_box_response


@pytest.fixture()
def reconnect_long_poll_response(make_mock_box_request):
    mock_box_response, _ = make_mock_box_request(
        response={'message': 'reconnect'},
    )
    return mock_box_response


@pytest.fixture()
def max_retries_long_poll_response(make_mock_box_request):
    mock_box_response, _ = make_mock_box_request(
        response={'message': 'max_retries'},
    )
    return mock_box_response


@pytest.fixture()
def mock_event():
    return {
        "type": "event",
        "event_id": "f82c3ba03e41f7e8a7608363cc6c0390183c3f83",
        "source": {
            "type": "folder",
            "id": "11446498",
        }
    }


@pytest.fixture()
def events_response(initial_stream_position, mock_event, make_mock_box_request):
    # pylint:disable=redefined-outer-name
    mock_box_response, _ = make_mock_box_request(
        response={"next_stream_position": initial_stream_position, "entries": [mock_event]},
    )
    return mock_box_response


def test_get_events(test_events, mock_box_session, events_response):
    # pylint:disable=redefined-outer-name
    mock_box_session.get.return_value = events_response
    events = test_events.get_events()
    assert 'next_stream_position' in events


def test_generate_events_with_long_polling(
        test_events,
        mock_box_session,
        events_response,
        empty_events_response,
        initial_stream_position,
        long_poll_url,
        retry_timeout,
        options_response,
        new_change_long_poll_response,
        reconnect_long_poll_response,
        max_retries_long_poll_response,
        mock_event,
):
    # pylint:disable=redefined-outer-name
    expected_url = test_events.get_url()
    mock_box_session.options.return_value = options_response
    mock_box_session.get.side_effect = [
        events_response,  # initial call to get now stream position
        TimeoutError,
        reconnect_long_poll_response,
        max_retries_long_poll_response,
        new_change_long_poll_response,
        events_response,
        new_change_long_poll_response,
        empty_events_response,
    ]
    events = test_events.generate_events_with_long_polling()
    assert next(events) == mock_event
    with pytest.raises(StopIteration):
        next(events)
    events.close()
    mock_box_session.options.assert_called_with(expected_url)
    mock_box_session.get.assert_any_call(expected_url, params={'stream_position': 'now'})
    assert '/events' in expected_url
    mock_box_session.get.assert_any_call(
        expected_url,
        params={'limit': 100, 'stream_type': 'all', 'stream_position': initial_stream_position},
    )
    mock_box_session.get.assert_any_call(
        long_poll_url,
        timeout=retry_timeout,
        params={'stream_position': initial_stream_position},
    )
